import logging
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from typing import TYPE_CHECKING, AbstractSet, Optional, Union, cast  # noqa: UP035

import dagster._check as check
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_graph_subset import AssetGraphSubset
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._core.definitions.data_version import DataVersion, extract_data_version_from_entry
from dagster._core.definitions.declarative_automation.legacy.valid_asset_subset import (
    ValidAssetSubset,
)
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.partition import PartitionsDefinition, PartitionsSubset
from dagster._core.definitions.time_window_partitions import (
    TimeWindowPartitionsDefinition,
    get_time_partition_key,
    get_time_partitions_def,
)
from dagster._core.errors import (
    DagsterDefinitionChangedDeserializationError,
    DagsterInvalidDefinitionError,
)
from dagster._core.event_api import AssetRecordsFilter
from dagster._core.events import DagsterEventType
from dagster._core.instance import DagsterInstance, DynamicPartitionsStore
from dagster._core.loader import LoadingContext
from dagster._core.storage.dagster_run import (
    IN_PROGRESS_RUN_STATUSES,
    DagsterRun,
    DagsterRunStatus,
    RunRecord,
)
from dagster._core.storage.tags import PARTITION_NAME_TAG
from dagster._time import get_current_datetime
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.execution.asset_backfill import AssetBackfillData
    from dagster._core.storage.event_log import EventLogRecord
    from dagster._core.storage.event_log.base import AssetRecord
    from dagster._core.storage.partition_status_cache import AssetStatusCacheValue

RECORD_BATCH_SIZE = 1000


class CachingInstanceQueryer(DynamicPartitionsStore):
    """Provides utility functions for querying for asset-materialization related data from the
    instance which will attempt to limit redundant expensive calls. Intended for use within the
    scope of a single "request" (e.g. GQL request, sensor tick).

    Args:
        instance (DagsterInstance): The instance to query.
    """

    def __init__(
        self,
        instance: DagsterInstance,
        asset_graph: BaseAssetGraph,
        loading_context: LoadingContext,
        evaluation_time: Optional[datetime] = None,
        logger: Optional[logging.Logger] = None,
    ):
        self._instance = instance
        self._loading_context = loading_context

        self._asset_graph = asset_graph
        self._logger = logger or logging.getLogger("dagster")

        self._asset_partitions_cache: dict[Optional[int], dict[AssetKey, set[str]]] = defaultdict(
            dict
        )
        self._asset_partition_versions_updated_after_cursor_cache: dict[
            AssetKeyPartitionKey, int
        ] = {}

        self._dynamic_partitions_cache: dict[str, Sequence[str]] = {}

        self._evaluation_time = evaluation_time if evaluation_time else get_current_datetime()

        self._respect_materialization_data_versions = (
            self._instance.auto_materialize_respect_materialization_data_versions
        )

    @property
    def instance(self) -> DagsterInstance:
        return self._instance

    @property
    def asset_graph(self) -> BaseAssetGraph:
        return self._asset_graph

    @property
    def evaluation_time(self) -> datetime:
        return self._evaluation_time

    ####################
    # QUERY BATCHING
    ####################

    def prefetch_asset_records(self, asset_keys: Iterable[AssetKey]):
        """For performance, batches together queries for selected assets."""
        from dagster._core.storage.event_log.base import AssetRecord

        AssetRecord.blocking_get_many(self._loading_context, asset_keys)

    ####################
    # ASSET STATUS CACHE
    ####################

    @cached_method
    def _get_updated_cache_value(self, *, asset_key: AssetKey) -> Optional["AssetStatusCacheValue"]:
        from dagster._core.storage.partition_status_cache import (
            get_and_update_asset_status_cache_value,
        )

        partitions_def = check.not_none(self.asset_graph.get(asset_key).partitions_def)
        return get_and_update_asset_status_cache_value(
            instance=self.instance,
            asset_key=asset_key,
            partitions_def=partitions_def,
            dynamic_partitions_loader=self,
            loading_context=self._loading_context,
        )

    @cached_method
    def get_failed_or_in_progress_subset(self, *, asset_key: AssetKey) -> PartitionsSubset:
        """Returns a PartitionsSubset representing the set of partitions that are either in progress
        or whose last materialization attempt failed.
        """
        partitions_def = check.not_none(self.asset_graph.get(asset_key).partitions_def)
        cache_value = self._get_updated_cache_value(asset_key=asset_key)
        if cache_value is None:
            return partitions_def.empty_subset()

        return cache_value.deserialize_failed_partition_subsets(
            partitions_def
        ) | cache_value.deserialize_in_progress_partition_subsets(partitions_def)

    @cached_method
    def get_materialized_asset_subset(
        self, *, asset_key: AssetKey
    ) -> SerializableEntitySubset[AssetKey]:
        """Returns an AssetSubset representing the subset of the asset that has been materialized."""
        partitions_def = self.asset_graph.get(asset_key).partitions_def
        if partitions_def:
            cache_value = self._get_updated_cache_value(asset_key=asset_key)
            if cache_value is None:
                value = partitions_def.empty_subset()
            else:
                value = cache_value.deserialize_materialized_partition_subsets(partitions_def)
        else:
            value = self.asset_partition_has_materialization_or_observation(
                AssetKeyPartitionKey(asset_key)
            )
        return SerializableEntitySubset(key=asset_key, value=value)

    @cached_method
    def get_in_progress_asset_subset(
        self, *, asset_key: AssetKey
    ) -> SerializableEntitySubset[AssetKey]:
        """Returns an AssetSubset representing the subset of the asset that is currently in progress."""
        partitions_def = self.asset_graph.get(asset_key).partitions_def
        if partitions_def:
            cache_value = self._get_updated_cache_value(asset_key=asset_key)
            if cache_value is None:
                value = partitions_def.empty_subset()
            else:
                value = cache_value.deserialize_in_progress_partition_subsets(partitions_def)
        else:
            # NOTE: this computation is not correct in all cases for unpartitioned assets. it is
            # possible (though rare) for run A to be launched targeting an asset, then later run B
            # be launched, and then run B completes before run A. In these cases, the computation
            # below will consider the asset to not be in progress, as the latest planned event
            # will be associated with a completed run.
            asset_record = self.get_asset_record(asset_key)
            last_materialized_run_id = (
                asset_record.asset_entry.last_materialization_record.run_id
                if asset_record and asset_record.asset_entry.last_materialization_record
                else None
            )

            planned_materialization_run_id = None
            if self.instance.event_log_storage.asset_records_have_last_planned_materialization_storage_id:
                planned_materialization_run_id = (
                    asset_record.asset_entry.last_planned_materialization_run_id
                    if asset_record
                    else None
                )
            else:
                planned_materialization_info = (
                    self.instance.event_log_storage.get_latest_planned_materialization_info(
                        asset_key
                    )
                )
                planned_materialization_run_id = (
                    planned_materialization_info.run_id if planned_materialization_info else None
                )
            if (
                not planned_materialization_run_id
                # if the latest materialization happened in the same run as the latest planned materialization,
                # it's no longer in progress
                or last_materialized_run_id == planned_materialization_run_id
            ):
                value = False
            else:
                dagster_run = self.instance.get_run_by_id(planned_materialization_run_id)
                value = dagster_run is not None and dagster_run.status in [
                    *IN_PROGRESS_RUN_STATUSES,
                    # an asset is considered to be "in progress" if there is planned work for it that has not
                    # yet completed, which is not identical to the "in progress" status of the run
                    DagsterRunStatus.QUEUED,
                ]

        return SerializableEntitySubset(key=asset_key, value=value)

    @cached_method
    def get_failed_asset_subset(self, *, asset_key: AssetKey) -> SerializableEntitySubset[AssetKey]:
        """Returns an AssetSubset representing the subset of the asset that failed to be
        materialized its most recent run.
        """
        partitions_def = self.asset_graph.get(asset_key).partitions_def
        if partitions_def:
            cache_value = self._get_updated_cache_value(asset_key=asset_key)
            if cache_value is None:
                value = partitions_def.empty_subset()
            else:
                value = cache_value.deserialize_failed_partition_subsets(partitions_def)
        else:
            # ideally, unpartitioned assets would also be handled by the asset status cache
            planned_materialization_info = (
                self.instance.event_log_storage.get_latest_planned_materialization_info(asset_key)
            )
            if not planned_materialization_info:
                value = False
            else:
                dagster_run = self.instance.get_run_by_id(planned_materialization_info.run_id)

                value = dagster_run is not None and dagster_run.status == DagsterRunStatus.FAILURE

        return SerializableEntitySubset(key=asset_key, value=value)

    ####################
    # ASSET RECORDS / STORAGE IDS
    ####################

    def get_asset_record(self, asset_key: AssetKey) -> Optional["AssetRecord"]:
        from dagster._core.storage.event_log.base import AssetRecord

        return AssetRecord.blocking_get(self._loading_context, asset_key)

    def _event_type_for_key(self, asset_key: AssetKey) -> DagsterEventType:
        if self.asset_graph.get(asset_key).is_observable:
            return DagsterEventType.ASSET_OBSERVATION
        else:
            return DagsterEventType.ASSET_MATERIALIZATION

    @cached_method
    def _get_latest_materialization_or_observation_record(
        self, *, asset_partition: AssetKeyPartitionKey, before_cursor: Optional[int] = None
    ) -> Optional["EventLogRecord"]:
        """Returns the latest event log record for the given asset partition of an asset. For
        observable source assets, this will be an AssetObservation, otherwise it will be an
        AssetMaterialization.
        """
        # in the simple case, just use the asset record
        if (
            before_cursor is None
            and asset_partition.partition_key is None
            and not (
                self.asset_graph.has(asset_partition.asset_key)
                and self.asset_graph.get(asset_partition.asset_key).is_observable
            )
        ):
            asset_record = self.get_asset_record(asset_partition.asset_key)
            if asset_record is None:
                return None
            return asset_record.asset_entry.last_materialization_record

        records_filter = AssetRecordsFilter(
            asset_key=asset_partition.asset_key,
            asset_partitions=(
                [asset_partition.partition_key] if asset_partition.partition_key else None
            ),
            before_storage_id=before_cursor,
        )

        # For observable assets, we fetch the most recent observation and materialization and return
        # whichever is more recent. For non-observable assets, we just fetch the most recent
        # materialization.
        materialization_records = self.instance.fetch_materializations(
            records_filter, ascending=False, limit=1
        ).records
        if self.asset_graph.get(asset_partition.asset_key).is_observable:
            observation_records = self.instance.fetch_observations(
                records_filter, ascending=False, limit=1
            ).records
            all_records = sorted(
                [*materialization_records, *observation_records],
                key=lambda x: x.timestamp,
                reverse=True,
            )
        else:
            all_records = materialization_records
        return next(iter(all_records), None)

    @cached_method
    def _get_latest_materialization_or_observation_storage_ids_by_asset_partition(
        self, *, asset_key: AssetKey
    ) -> Mapping[AssetKeyPartitionKey, Optional[int]]:
        """Returns a mapping from asset partition to the latest storage id for that asset partition
        for all asset partitions associated with the given asset key.

        Note that for partitioned assets, an asset partition with a None partition key will be
        present in the mapping, representing the latest storage id for the asset as a whole.
        """
        asset_partition = AssetKeyPartitionKey(asset_key)
        latest_record = self._get_latest_materialization_or_observation_record(
            asset_partition=asset_partition
        )
        latest_storage_ids = {
            asset_partition: latest_record.storage_id if latest_record is not None else None
        }
        if self.asset_graph.get(asset_key).is_partitioned:
            latest_storage_ids.update(
                {
                    AssetKeyPartitionKey(asset_key, partition_key): storage_id
                    for partition_key, storage_id in self.instance.get_latest_storage_id_by_partition(
                        asset_key, event_type=self._event_type_for_key(asset_key)
                    ).items()
                }
            )
        return latest_storage_ids

    def get_latest_materialization_or_observation_storage_id(
        self, asset_partition: AssetKeyPartitionKey
    ) -> Optional[int]:
        """Returns the latest storage id for the given asset partition. If the asset has never been
        materialized, returns None.

        Args:
            asset_partition (AssetKeyPartitionKey): The asset partition to query.
        """
        if asset_partition.partition_key is None:
            record = self._get_latest_materialization_or_observation_record(
                asset_partition=asset_partition
            )
            return record.storage_id if record else None
        return self._get_latest_materialization_or_observation_storage_ids_by_asset_partition(
            asset_key=asset_partition.asset_key
        ).get(asset_partition)

    def asset_partition_has_materialization_or_observation(
        self,
        asset_partition: AssetKeyPartitionKey,
        after_cursor: Optional[int] = None,
    ) -> bool:
        """Returns True if there is a materialization record for the given asset partition after
        the specified cursor.

        Args:
            asset_partition (AssetKeyPartitionKey): The asset partition to query.
            after_cursor (Optional[int]): Filter parameter such that only records with a storage_id
                greater than this value will be considered.
        """
        asset_key = asset_partition.asset_key
        if self.asset_graph.has(asset_key) and self.asset_graph.get(asset_key).is_materializable:
            asset_record = self.get_asset_record(asset_key)
            if (
                asset_record is None
                or asset_record.asset_entry.last_materialization_record is None
                or (
                    after_cursor
                    and asset_record.asset_entry.last_materialization_record.storage_id
                    <= after_cursor
                )
            ):
                return False
        return (self.get_latest_materialization_or_observation_storage_id(asset_partition) or 0) > (
            after_cursor or 0
        )

    def get_latest_materialization_or_observation_record(
        self,
        asset_partition: AssetKeyPartitionKey,
        after_cursor: Optional[int] = None,
        before_cursor: Optional[int] = None,
    ) -> Optional["EventLogRecord"]:
        """Returns the latest record for the given asset partition given the specified cursors.

        Args:
            asset_partition (AssetKeyPartitionKey): The asset partition to query.
            after_cursor (Optional[int]): Filter parameter such that only records with a storage_id
                greater than this value will be considered.
            before_cursor (Optional[int]): Filter parameter such that only records with a storage_id
                less than this value will be considered.
        """
        check.param_invariant(
            not (after_cursor and before_cursor),
            "before_cursor",
            "Cannot set both before_cursor and after_cursor",
        )

        # first, do a quick check to eliminate the case where we know there is no record
        if not self.asset_partition_has_materialization_or_observation(
            asset_partition, after_cursor
        ):
            return None
        # then, if the before_cursor is after our latest record's storage id, we can just return
        # the latest record
        elif (before_cursor or 0) > (
            self.get_latest_materialization_or_observation_storage_id(asset_partition) or 0
        ):
            return self._get_latest_materialization_or_observation_record(
                asset_partition=asset_partition
            )
        # otherwise, do the explicit query
        return self._get_latest_materialization_or_observation_record(
            asset_partition=asset_partition, before_cursor=before_cursor
        )

    ####################
    # OBSERVATIONS
    ####################

    @cached_method
    def next_version_record(
        self,
        *,
        asset_key: AssetKey,
        after_cursor: Optional[int],
        data_version: Optional[DataVersion],
    ) -> Optional["EventLogRecord"]:
        has_more = True
        cursor = None
        while has_more:
            result = self.instance.fetch_observations(
                AssetRecordsFilter(asset_key=asset_key, after_storage_id=after_cursor),
                limit=RECORD_BATCH_SIZE,
                cursor=cursor,
                ascending=True,
            )
            has_more = result.has_more
            cursor = result.cursor
            for record in result.records:
                record_version = extract_data_version_from_entry(record.event_log_entry)
                if record_version is not None and record_version != data_version:
                    return record

        # no records found with a new data version
        return None

    ####################
    # RUNS
    ####################

    @cached_method
    def _get_run_record_by_id(self, *, run_id: str) -> Optional[RunRecord]:
        return self.instance.get_run_record_by_id(run_id)

    def _get_run_by_id(self, run_id: str) -> Optional[DagsterRun]:
        run_record = self._get_run_record_by_id(run_id=run_id)
        if run_record is not None:
            return run_record.dagster_run
        return None

    def run_has_tag(self, run_id: str, tag_key: str, tag_value: Optional[str]) -> bool:
        run = self._get_run_by_id(run_id)
        if run is None:
            return False

        run_tags = run.tags

        if tag_value is None:
            return tag_key in run_tags
        else:
            return run_tags.get(tag_key) == tag_value

    @cached_method
    def _get_planned_materializations_for_run_from_snapshot(
        self, *, snapshot_id: str
    ) -> AbstractSet[AssetKey]:
        execution_plan_snapshot = check.not_none(
            self._instance.get_execution_plan_snapshot(snapshot_id)
        )
        return execution_plan_snapshot.asset_selection

    @cached_method
    def _get_planned_materializations_for_run_from_events(
        self, *, run_id: str
    ) -> AbstractSet[AssetKey]:
        """Provides a fallback for fetching the planned materializations for a run from
        the ASSET_MATERIALIZATION_PLANNED events in the event log, in cases where this information
        is not available on the DagsterRun object.

        Args:
            run_id (str): The run id
        """
        materializations_planned = self.instance.get_records_for_run(
            run_id=run_id, of_type=DagsterEventType.ASSET_MATERIALIZATION_PLANNED
        ).records
        return set(cast(AssetKey, record.asset_key) for record in materializations_planned)

    def get_planned_materializations_for_run(self, run_id: str) -> AbstractSet[AssetKey]:
        """Returns the set of asset keys that are planned to be materialized by the run.

        Args:
            run_id (str): The run id
        """
        run = self._get_run_by_id(run_id=run_id)
        if run is None:
            return set()
        elif run.asset_selection:
            return run.asset_selection
        elif run.execution_plan_snapshot_id:
            return self._get_planned_materializations_for_run_from_snapshot(
                snapshot_id=check.not_none(run.execution_plan_snapshot_id)
            )
        else:
            # must resort to querying the event log
            return self._get_planned_materializations_for_run_from_events(run_id=run_id)

    def is_asset_planned_for_run(
        self, run_id: str, asset: Union[AssetKey, AssetKeyPartitionKey]
    ) -> bool:
        """Returns True if the asset is planned to be materialized by the run."""
        run = self._get_run_by_id(run_id=run_id)
        if not run:
            return False

        if isinstance(asset, AssetKeyPartitionKey):
            asset_key = asset.asset_key
            if run.tags.get(PARTITION_NAME_TAG) != asset.partition_key:
                return False
        else:
            asset_key = asset

        return asset_key in self.get_planned_materializations_for_run(run_id=run_id)

    @cached_method
    def get_current_materializations_for_run(self, *, run_id: str) -> AbstractSet[AssetKey]:
        """Returns the set of asset keys that have been materialized by a given run.

        Args:
            run_id (str): The run id
        """
        materializations = self.instance.get_records_for_run(
            run_id=run_id,
            of_type=DagsterEventType.ASSET_MATERIALIZATION,
        ).records
        return set(cast(AssetKey, record.asset_key) for record in materializations)

    ####################
    # BACKFILLS
    ####################

    @cached_method
    def get_active_backfill_datas(self) -> Sequence["AssetBackfillData"]:
        from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus

        active_backfills = [
            backfill
            for backfill in self.instance.get_backfills(
                filters=BulkActionsFilter(statuses=[BulkActionStatus.REQUESTED])
            )
            if backfill.is_asset_backfill
        ]

        backfill_datas = []
        for backfill in active_backfills:
            try:
                backfill_datas.append(backfill.get_asset_backfill_data(self.asset_graph))
            except DagsterDefinitionChangedDeserializationError:
                self._logger.warning(
                    f"Not considering assets in backfill {backfill.backfill_id} since its"
                    " data could not be deserialized"
                )
                # Backfill can't be loaded, so no risk of the assets interfering
                continue
        return backfill_datas

    @cached_method
    def get_active_backfill_target_asset_graph_subset(self) -> AssetGraphSubset:
        """Returns an AssetGraphSubset representing the set of assets that are currently targeted by
        an active asset backfill.
        """
        result = AssetGraphSubset()
        for data in self.get_active_backfill_datas():
            result |= data.target_subset

        return result

    @cached_method
    def get_active_backfill_in_progress_asset_graph_subset(self) -> AssetGraphSubset:
        """Returns an AssetGraphSubset representing the set of assets that are currently targeted by
        an active asset backfill and have not yet been materialized or failed.
        """
        result = AssetGraphSubset()
        for data in self.get_active_backfill_datas():
            in_progress_subset = (
                data.target_subset - data.materialized_subset - data.failed_and_downstream_subset
            )
            result |= in_progress_subset

        return result

    ####################
    # PARTITIONS
    ####################

    def get_materialized_partitions(
        self, asset_key: AssetKey, before_cursor: Optional[int] = None
    ) -> set[str]:
        """Returns a list of the partitions that have been materialized for the given asset key.

        Args:
            asset_key (AssetKey): The asset key.
            before_cursor (Optional[int]): The cursor before which to look for materialized
                partitions. If not provided, will look at all materializations.
        """
        if (
            before_cursor not in self._asset_partitions_cache
            or asset_key not in self._asset_partitions_cache[before_cursor]
        ):
            self._asset_partitions_cache[before_cursor][asset_key] = (
                self.instance.get_materialized_partitions(
                    asset_key=asset_key, before_cursor=before_cursor
                )
            )

        return self._asset_partitions_cache[before_cursor][asset_key]

    ####################
    # DYNAMIC PARTITIONS
    ####################

    def get_dynamic_partitions(self, partitions_def_name: str) -> Sequence[str]:
        """Returns a list of partitions for a partitions definition."""
        if partitions_def_name not in self._dynamic_partitions_cache:
            self._dynamic_partitions_cache[partitions_def_name] = (
                self.instance.get_dynamic_partitions(partitions_def_name)
            )
        return self._dynamic_partitions_cache[partitions_def_name]

    def has_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> bool:
        return partition_key in self.get_dynamic_partitions(partitions_def_name)

    @cached_method
    def asset_partitions_with_newly_updated_parents_and_new_cursor(
        self,
        *,
        latest_storage_id: Optional[int],
        child_asset_key: AssetKey,
        map_old_time_partitions: bool = True,
        max_child_partitions: Optional[int] = None,
    ) -> tuple[AbstractSet[AssetKeyPartitionKey], Optional[int]]:
        """Finds asset partitions of the given child whose parents have been materialized since
        latest_storage_id.
        """
        max_storage_ids = [
            self.get_latest_materialization_or_observation_storage_id(
                AssetKeyPartitionKey(child_asset_key)
            )
        ]
        child_asset = self.asset_graph.get(child_asset_key)
        if not child_asset.parent_keys:
            return set(), max(filter(None, [latest_storage_id, *max_storage_ids]), default=None)

        child_time_partitions_def = get_time_partitions_def(child_asset.partitions_def)
        child_asset_partitions_with_updated_parents = set()
        for parent_asset_key in self.asset_graph.get(child_asset_key).parent_keys:
            # ignore non-existent parents
            if not self.asset_graph.has(parent_asset_key):
                continue

            # if the parent has not been updated at all since the latest_storage_id, then skip
            if not self.get_asset_partitions_updated_after_cursor(
                asset_key=parent_asset_key,
                asset_partitions=None,
                after_cursor=latest_storage_id,
                respect_materialization_data_versions=False,
            ):
                continue

            # keep track of the maximum storage id that we've seen for a given parent
            max_storage_ids.append(
                self.get_latest_materialization_or_observation_storage_id(
                    AssetKeyPartitionKey(parent_asset_key)
                )
            )

            parent_partitions_def: PartitionsDefinition = self.asset_graph.get(
                parent_asset_key
            ).partitions_def
            if parent_partitions_def is None:
                latest_parent_record = check.not_none(
                    self.get_latest_materialization_or_observation_record(
                        AssetKeyPartitionKey(parent_asset_key), after_cursor=latest_storage_id
                    )
                )
                for child_partition_key in (
                    self.asset_graph.get_child_partition_keys_of_parent(
                        dynamic_partitions_store=self,
                        parent_partition_key=None,
                        parent_asset_key=parent_asset_key,
                        child_asset_key=child_asset_key,
                        current_time=self.evaluation_time,
                    )
                    if child_asset.partitions_def
                    else [None]
                ):
                    if not (
                        # when mapping from unpartitioned assets to time partitioned assets, we ignore
                        # historical time partitions
                        not map_old_time_partitions
                        and child_time_partitions_def is not None
                        and get_time_partition_key(child_asset.partitions_def, child_partition_key)
                        != child_time_partitions_def.get_last_partition_key(
                            current_time=self.evaluation_time
                        )
                    ) and not self.is_asset_planned_for_run(
                        latest_parent_record.run_id, child_asset_key
                    ):
                        child_asset_partitions_with_updated_parents.add(
                            AssetKeyPartitionKey(child_asset_key, child_partition_key)
                        )
            else:
                # we know a parent updated, and because the parent has a partitions def and the
                # child does not, the child could not have been materialized in the same run
                if child_asset.partitions_def is None:
                    child_asset_partitions_with_updated_parents = {
                        AssetKeyPartitionKey(child_asset_key)
                    }
                    break

                # the set of asset partitions which have been updated since the latest storage id
                parent_partitions_subset = self.get_asset_subset_updated_after_cursor(
                    asset_key=parent_asset_key,
                    after_cursor=latest_storage_id,
                    require_data_version_update=False,
                ).subset_value

                # we are mapping from the partitions of the parent asset to the partitions of
                # the child asset
                partition_mapping = self.asset_graph.get_partition_mapping(
                    key=child_asset_key, parent_asset_key=parent_asset_key
                )
                try:
                    child_partitions_subset = (
                        partition_mapping.get_downstream_partitions_for_partitions(
                            parent_partitions_subset,
                            upstream_partitions_def=parent_partitions_def,
                            downstream_partitions_def=child_asset.partitions_def,
                            dynamic_partitions_store=self,
                            current_time=self.evaluation_time,
                        )
                    )
                except DagsterInvalidDefinitionError as e:
                    # add a more helpful error message to the stack
                    raise DagsterInvalidDefinitionError(
                        f"Could not map partitions between parent {parent_asset_key.to_string()} "
                        f"and child {child_asset_key.to_string()}."
                    ) from e

                # Prefer more recent time-based partitions, particularly if we end up filtering
                # using max_child_partitions (not a strict guarantee that this will always return
                # the most recent partitions in time though)
                child_partitions = sorted(
                    child_partitions_subset.get_partition_keys(), reverse=True
                )

                if max_child_partitions is not None:
                    child_partitions = child_partitions[:max_child_partitions]

                for child_partition in child_partitions:
                    # we need to see if the child is planned for the same run, but this is
                    # expensive, so we try to avoid doing so in as many situations as possible
                    child_asset_partition = AssetKeyPartitionKey(child_asset_key, child_partition)
                    if (
                        # if child has a different partitions def than the parent, then it must
                        # have been executed in a different run, so it's a valid candidate
                        child_asset.partitions_def != parent_partitions_def
                        # if child partition key is not the same as any newly materialized
                        # parent key, then it could not have been executed in the same run as
                        # its parent
                        or child_partition not in parent_partitions_subset
                        # if child partition is not failed or in progress, then if it was
                        # executed in the same run as its parent, then it must have been
                        # materialized more recently than its parent
                        or child_partition
                        not in self.get_failed_or_in_progress_subset(asset_key=child_asset_key)
                    ):
                        child_asset_partitions_with_updated_parents.add(child_asset_partition)
                    else:
                        # manually query to see if this asset partition was intended to be
                        # executed in the same run as its parent
                        latest_partition_record = check.not_none(
                            self.get_latest_materialization_or_observation_record(
                                AssetKeyPartitionKey(parent_asset_key, child_partition),
                                after_cursor=latest_storage_id,
                            )
                        )
                        if not self.is_asset_planned_for_run(
                            latest_partition_record.run_id, child_asset_key
                        ):
                            child_asset_partitions_with_updated_parents.add(child_asset_partition)

        # the new latest storage id will be the greatest observed storage id among this asset and
        # its parents
        new_latest_storage_id = max(
            filter(None, [latest_storage_id, *max_storage_ids]), default=None
        )
        return (child_asset_partitions_with_updated_parents, new_latest_storage_id)

    ####################
    # RECONCILIATION
    ####################

    def _asset_partition_versions_updated_after_cursor(
        self,
        asset_key: AssetKey,
        asset_partitions: AbstractSet[AssetKeyPartitionKey],
        after_cursor: int,
    ) -> AbstractSet[AssetKeyPartitionKey]:
        # we already know asset partitions are updated after the cursor if they've been updated
        # after a cursor that's greater than or equal to this one
        updated_asset_partitions = {
            ap
            for ap in asset_partitions
            if ap in self._asset_partition_versions_updated_after_cursor_cache
            and self._asset_partition_versions_updated_after_cursor_cache[ap] <= after_cursor
        }
        to_query_asset_partitions = asset_partitions - updated_asset_partitions
        if not to_query_asset_partitions:
            return updated_asset_partitions

        if not self.asset_graph.get(asset_key).is_partitioned:
            asset_partition = AssetKeyPartitionKey(asset_key)
            latest_record = self.get_latest_materialization_or_observation_record(
                asset_partition, after_cursor=after_cursor
            )
            latest_data_version = (
                extract_data_version_from_entry(latest_record.event_log_entry)
                if latest_record
                else None
            )
            previous_record = self.get_latest_materialization_or_observation_record(
                asset_partition, before_cursor=after_cursor + 1
            )
            previous_data_version = (
                extract_data_version_from_entry(previous_record.event_log_entry)
                if previous_record
                else None
            )
            return set([asset_partition]) if latest_data_version != previous_data_version else set()

        partition_keys = [
            asset_partition.partition_key
            for asset_partition in asset_partitions
            if asset_partition.partition_key
        ]
        updated_partition_keys = (
            self.instance.event_log_storage.get_updated_data_version_partitions(
                asset_key=asset_key,
                partitions=partition_keys,
                since_storage_id=after_cursor,
            )
        )
        return set(
            AssetKeyPartitionKey(asset_key, partition_key)
            for partition_key in updated_partition_keys
        )

    def get_asset_partitions_updated_after_cursor(
        self,
        asset_key: AssetKey,
        asset_partitions: Optional[AbstractSet[AssetKeyPartitionKey]],
        after_cursor: Optional[int],
        respect_materialization_data_versions: bool,
    ) -> AbstractSet[AssetKeyPartitionKey]:
        """Returns the set of asset partitions that have been updated after the given cursor.

        Args:
            asset_key (AssetKey): The asset key to check.
            asset_partitions (Optional[Sequence[AssetKeyPartitionKey]]): If supplied, will filter
                the set of checked partitions to the given partitions.
            after_cursor (Optional[int]): The cursor after which to look for updates.
            respect_materialization_data_versions (bool): If True, will use data versions to filter
                out asset partitions which were materialized, but not have not had their data
                versions changed since the given cursor.
                NOTE: This boolean has been temporarily disabled
        """
        if not self.asset_partition_has_materialization_or_observation(
            AssetKeyPartitionKey(asset_key), after_cursor=after_cursor
        ):
            return set()

        last_storage_id_by_asset_partition = (
            self._get_latest_materialization_or_observation_storage_ids_by_asset_partition(
                asset_key=asset_key
            )
        )

        if asset_partitions is None:
            updated_after_cursor = {
                asset_partition
                for asset_partition, latest_storage_id in last_storage_id_by_asset_partition.items()
                if (latest_storage_id or 0) > (after_cursor or 0)
            }
        else:
            # Optimized for the case where there are many partitions and last_storage_id_by_asset_partition
            # is large, but we're only looking for the result for a small number of partitions
            updated_after_cursor = set()
            for asset_partition in asset_partitions:
                latest_storage_id = last_storage_id_by_asset_partition.get(asset_partition)
                if latest_storage_id is not None and latest_storage_id > (after_cursor or 0):
                    updated_after_cursor.add(asset_partition)

        if not updated_after_cursor:
            return set()
        if after_cursor is None or (
            not self.asset_graph.get(asset_key).is_observable
            and not respect_materialization_data_versions
        ):
            return updated_after_cursor

        return self._asset_partition_versions_updated_after_cursor(
            asset_key, updated_after_cursor, after_cursor
        )

    @cached_method
    def get_asset_subset_updated_after_cursor(
        self, *, asset_key: AssetKey, after_cursor: Optional[int], require_data_version_update: bool
    ) -> SerializableEntitySubset[AssetKey]:
        """Returns the AssetSubset of the given asset that has been updated after the given cursor."""
        partitions_def = self.asset_graph.get(asset_key).partitions_def
        updated_asset_partitions = self.get_asset_partitions_updated_after_cursor(
            asset_key,
            asset_partitions=None,
            after_cursor=after_cursor,
            respect_materialization_data_versions=require_data_version_update,
        )
        if partitions_def is None:
            validated_asset_partitions = {
                ap for ap in updated_asset_partitions if ap.partition_key is None
            }
        else:
            validated_asset_partitions = {
                ap
                for ap in updated_asset_partitions
                if ap.partition_key is not None
                and partitions_def.has_partition_key(
                    partition_key=ap.partition_key,
                    dynamic_partitions_store=self,
                    current_time=self.evaluation_time,
                )
            }

        # TODO: replace this return value with EntitySubset
        return ValidAssetSubset.from_asset_partitions_set(
            asset_key, partitions_def, validated_asset_partitions
        )

    @cached_method
    def get_asset_subset_updated_after_time(
        self, *, asset_key: AssetKey, after_time: datetime
    ) -> SerializableEntitySubset[AssetKey]:
        """Returns the AssetSubset of the given asset that has been updated after the given time."""
        partitions_def = self.asset_graph.get(asset_key).partitions_def

        method = (
            self.instance.fetch_materializations
            if self._event_type_for_key(asset_key) == DagsterEventType.ASSET_MATERIALIZATION
            else self.instance.fetch_observations
        )
        first_event_after_time = next(
            iter(
                method(
                    AssetRecordsFilter(asset_key=asset_key, after_timestamp=after_time.timestamp()),
                    limit=1,
                    ascending=True,
                ).records
            ),
            None,
        )
        if not first_event_after_time:
            # TODO: replace this return value with EntitySubset
            return ValidAssetSubset.empty(asset_key, partitions_def=partitions_def)
        else:
            return self.get_asset_subset_updated_after_cursor(
                asset_key=asset_key,
                after_cursor=first_event_after_time.storage_id - 1,
                require_data_version_update=False,
            )

    def get_parent_asset_partitions_updated_after_child(
        self,
        *,
        asset_partition: AssetKeyPartitionKey,
        parent_asset_partitions: AbstractSet[AssetKeyPartitionKey],
        respect_materialization_data_versions: bool,
        ignored_parent_keys: AbstractSet[AssetKey],
    ) -> AbstractSet[AssetKeyPartitionKey]:
        """Returns values inside parent_asset_partitions that correspond to asset partitions that
        have been updated since the latest materialization of asset_partition.
        """
        parent_asset_partitions_by_key: dict[AssetKey, set[AssetKeyPartitionKey]] = defaultdict(set)
        for parent in parent_asset_partitions:
            parent_asset_partitions_by_key[parent.asset_key].add(parent)

        partitions_def = self.asset_graph.get(asset_partition.asset_key).partitions_def
        updated_parents = set()

        for parent_key, parent_asset_partitions in parent_asset_partitions_by_key.items():
            # ignore updates to particular parents
            if parent_key in ignored_parent_keys:
                continue

            # ignore non-existent parents
            if not self.asset_graph.has(parent_key):
                continue

            # when mapping from unpartitioned assets to time partitioned assets, we ignore
            # historical time partitions
            if (
                isinstance(partitions_def, TimeWindowPartitionsDefinition)
                and not self.asset_graph.get(parent_key).is_partitioned
                and asset_partition.partition_key
                != partitions_def.get_last_partition_key(
                    current_time=self.evaluation_time, dynamic_partitions_store=self
                )
            ):
                continue

            updated_parents.update(
                self.get_asset_partitions_updated_after_cursor(
                    asset_key=parent_key,
                    asset_partitions=parent_asset_partitions,
                    after_cursor=self.get_latest_materialization_or_observation_storage_id(
                        asset_partition
                    ),
                    respect_materialization_data_versions=respect_materialization_data_versions,
                )
            )
        return updated_parents

    def have_ignorable_partition_mapping_for_outdated(
        self, asset_key: AssetKey, upstream_asset_key: AssetKey
    ) -> bool:
        """Returns whether the given assets have a partition mapping between them which can be
        ignored in the context of calculating if an asset is outdated or not.

        These mappings are ignored in cases where respecting them would require an unrealistic
        number of upstream partitions to be in a 'good' state before allowing a downstream asset
        to be considered up to date.
        """
        # Self partition mappings impose constraints on all historical partitions
        return asset_key == upstream_asset_key

    @cached_method
    def get_outdated_ancestors(
        self, *, asset_partition: AssetKeyPartitionKey
    ) -> AbstractSet[AssetKey]:
        asset_key = asset_partition.asset_key
        partition_key = asset_partition.partition_key
        if not (
            self.asset_graph.has(asset_key) and self.asset_graph.get(asset_key).is_materializable
        ):
            return set()

        parent_asset_partitions = self.asset_graph.get_parents_partitions(
            dynamic_partitions_store=self,
            current_time=self._evaluation_time,
            asset_key=asset_key,
            partition_key=partition_key,
        ).parent_partitions

        # the set of parent keys which we don't need to check
        ignored_parent_keys = {
            parent
            for parent in self.asset_graph.get(asset_key).parent_keys
            if self.have_ignorable_partition_mapping_for_outdated(asset_key, parent)
        }

        updated_parents = self.get_parent_asset_partitions_updated_after_child(
            asset_partition=asset_partition,
            parent_asset_partitions=parent_asset_partitions,
            respect_materialization_data_versions=self._respect_materialization_data_versions,
            ignored_parent_keys=ignored_parent_keys,
        )

        root_unreconciled_ancestors = {asset_key} if updated_parents else set()

        # recurse over parents
        for parent in set(parent_asset_partitions) - updated_parents:
            if parent.asset_key in ignored_parent_keys:
                continue
            root_unreconciled_ancestors.update(self.get_outdated_ancestors(asset_partition=parent))

        return root_unreconciled_ancestors



from typing import Any, Literal, Union

from typing_extensions import (
    Literal as ExtLiteral,
    get_origin,
)


def safe_is_subclass(cls: Any, possible_parent_cls: Union[type, tuple[type, ...]]) -> bool:
    """Version of issubclass that returns False if cls is not a Type."""
    if not isinstance(cls, type):
        return False

    try:
        return issubclass(cls, possible_parent_cls)
    except TypeError:
        # Using builtin Python types in python 3.9+ will raise a TypeError when using issubclass
        # even though the isinstance check will succeed (as will inspect.isclass), for example
        # list[dict[str, str]] will raise a TypeError
        return False


def is_literal(annotation: type) -> bool:
    return get_origin(annotation) in (Literal, ExtLiteral)



from dagster._core.storage.runs.base import RunStorage as RunStorage
from dagster._core.storage.runs.in_memory import InMemoryRunStorage as InMemoryRunStorage
from dagster._core.storage.runs.schema import (
    DaemonHeartbeatsTable as DaemonHeartbeatsTable,
    InstanceInfo as InstanceInfo,
    RunStorageSqlMetadata as RunStorageSqlMetadata,
)
from dagster._core.storage.runs.sql_run_storage import SqlRunStorage as SqlRunStorage
from dagster._core.storage.runs.sqlite import SqliteRunStorage as SqliteRunStorage



import inspect
from abc import ABC
from functools import update_wrapper
from types import MethodType

import dagster._check as check
from dagster._check.builder import (
    INJECTED_DEFAULT_VALS_LOCAL_VAR,
    EvalContext,
    build_args_and_assignment_strs,
    build_check_call_str,
)


class CheckedFnWrapper(ABC):
    """Output of the @checked decorator, this class holds a reference to the decorated function
    and upon first invocation compiles a wrapping function that performs run time type checks on
    annotated inputs.

    This class is not directly instantiated, but instead dynamic subclasses are created for each
    callsite allowing the __call__ method to be replaced[1] with the compiled function to achieve
    a single stack frame of over head for this decorator in best cases scenarios.


    [1] __call__ can not be replaced on instances, only classes.
    """

    def __init__(self, fn):
        self._target_fn = fn
        self._eval_ctx = EvalContext.capture_from_frame(2)

    def __get__(self, instance, _=None):
        """Allow the decorated function to be bound to instances to support class methods."""
        if instance:
            return MethodType(self, instance)
        return self

    def __call__(self, *args, **kwargs):
        signature = inspect.signature(self._target_fn)
        lines = []
        param_names = []
        defaults = {}
        for name, param in signature.parameters.items():
            if param.annotation != param.empty:
                param_str = build_check_call_str(
                    ttype=param.annotation,
                    name=name,
                    eval_ctx=self._eval_ctx,
                )
            else:
                param_str = param.name

            if param.kind in (param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD):
                param_str = f"{param.name}={param_str}"

            param_names.append(param.name)
            if param.default != param.empty:
                defaults[param.name] = param.default

            lines.append(param_str)

        lazy_imports_str = "\n    ".join(
            f"from {module} import {t}" for t, module in self._eval_ctx.lazy_imports.items()
        )
        args_str, set_calls_str = build_args_and_assignment_strs(
            param_names,
            defaults,
            kw_only=False,
        )
        param_block = ",\n        ".join(lines)

        checked_fn_name = f"__checked_{self._target_fn.__name__}"

        fn_str = f"""
def {checked_fn_name}(__checked_wrapper{args_str}):
    {lazy_imports_str}
    {set_calls_str}
    return __checked_wrapper._target_fn(
        {param_block}
    )
        """

        if "check" not in self._eval_ctx.global_ns:
            self._eval_ctx.global_ns["check"] = check

        self._eval_ctx.local_ns[INJECTED_DEFAULT_VALS_LOCAL_VAR] = defaults

        call = self._eval_ctx.compile_fn(
            fn_str,
            fn_name=checked_fn_name,
        )

        self.__class__.__call__ = call
        return call(self, *args, **kwargs)


def checked(fn):
    """Decorator for adding runtime type checking based on type annotations."""
    # if nothing can be checked, return the original fn
    annotations = getattr(fn, "__annotations__", None)
    if not annotations or (len(annotations) == 1 and set(annotations.keys()) == {"return"}):
        return fn

    # make a dynamic subclass to be able to hot swap __call__ post compilation
    class _DynamicCheckedFnWrapper(CheckedFnWrapper): ...

    checked_fn = _DynamicCheckedFnWrapper(fn)
    return update_wrapper(
        wrapper=checked_fn,
        wrapped=fn,
    )



from dagster._core.executor.step_delegating.step_handler.base import (
    CheckStepHealthResult as CheckStepHealthResult,
    StepHandler as StepHandler,
    StepHandlerContext as StepHandlerContext,
)



import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING, Union

from typing_extensions import TypeIs

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.declarative_automation.operators.dep_operators import (
    DepsAutomationCondition,
)
from dagster._record import copy, record
from dagster._serdes.serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.definitions.asset_selection import AssetSelection


def _has_allow_ignore(
    condition: AutomationCondition,
) -> TypeIs[
    Union[
        DepsAutomationCondition,
        "AndAutomationCondition",
        "OrAutomationCondition",
        "NotAutomationCondition",
    ]
]:
    from dagster._core.definitions.declarative_automation.operators.boolean_operators import (
        AndAutomationCondition,
        NotAutomationCondition,
        OrAutomationCondition,
    )

    return isinstance(
        condition,
        (
            DepsAutomationCondition,
            AndAutomationCondition,
            OrAutomationCondition,
            NotAutomationCondition,
        ),
    )


@whitelist_for_serdes(storage_name="AndAssetCondition")
@record
class AndAutomationCondition(BuiltinAutomationCondition[T_EntityKey]):
    """This class represents the condition that all of its children evaluate to true."""

    operands: Sequence[AutomationCondition[T_EntityKey]]

    @property
    def description(self) -> str:
        return "All of"

    @property
    def name(self) -> str:
        return "AND"

    @property
    def children(self) -> Sequence[AutomationCondition[T_EntityKey]]:
        return self.operands

    @property
    def requires_cursor(self) -> bool:
        return False

    async def evaluate(
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        child_results: list[AutomationResult] = []
        true_subset = context.candidate_subset
        for i, child in enumerate(self.children):
            child_result = await context.for_child_condition(
                child_condition=child, child_index=i, candidate_subset=true_subset
            ).evaluate_async()
            child_results.append(child_result)
            true_subset = true_subset.compute_intersection(child_result.true_subset)
        return AutomationResult(context, true_subset, child_results=child_results)

    def without(self, condition: AutomationCondition) -> "AndAutomationCondition":
        """Returns a copy of this condition without the specified child condition."""
        check.param_invariant(
            condition in self.operands, "condition", "Condition not found in operands"
        )
        operands = [child for child in self.operands if child != condition]
        if len(operands) < 2:
            check.failed("Cannot have fewer than 2 operands in an AndAutomationCondition")
        return copy(self, operands=[child for child in self.operands if child != condition])

    @public
    def replace(
        self, old: Union[AutomationCondition, str], new: AutomationCondition
    ) -> AutomationCondition:
        """Replaces all instances of ``old`` across any sub-conditions with ``new``.

        If ``old`` is a string, then conditions with a label matching
        that string will be replaced.

        Args:
            old (Union[AutomationCondition, str]): The condition to replace.
            new (AutomationCondition): The condition to replace with.
        """
        return (
            new
            if old in [self, self.get_label()]
            else copy(self, operands=[child.replace(old, new) for child in self.operands])
        )

    @public
    def allow(self, selection: "AssetSelection") -> "AndAutomationCondition":
        """Applies the ``.allow()`` method across all sub-conditions.

        This impacts any dep-related sub-conditions.

        Args:
            selection (AssetSelection): The selection to allow.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        return copy(
            self,
            operands=[
                child.allow(selection) if _has_allow_ignore(child) else child
                for child in self.operands
            ],
        )

    @public
    def ignore(self, selection: "AssetSelection") -> "AndAutomationCondition":
        """Applies the ``.ignore()`` method across all sub-conditions.

        This impacts any dep-related sub-conditions.

        Args:
            selection (AssetSelection): The selection to ignore.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        return copy(
            self,
            operands=[
                child.ignore(selection) if _has_allow_ignore(child) else child
                for child in self.operands
            ],
        )


@whitelist_for_serdes(storage_name="OrAssetCondition")
@record
class OrAutomationCondition(BuiltinAutomationCondition[T_EntityKey]):
    """This class represents the condition that any of its children evaluate to true."""

    operands: Sequence[AutomationCondition[T_EntityKey]]

    @property
    def description(self) -> str:
        return "Any of"

    @property
    def name(self) -> str:
        return "OR"

    @property
    def children(self) -> Sequence[AutomationCondition[T_EntityKey]]:
        return self.operands

    @property
    def requires_cursor(self) -> bool:
        return False

    async def evaluate(
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        true_subset = context.get_empty_subset()

        coroutines = [
            context.for_child_condition(
                child_condition=child, child_index=i, candidate_subset=context.candidate_subset
            ).evaluate_async()
            for i, child in enumerate(self.children)
        ]

        child_results = await asyncio.gather(*coroutines)
        for child_result in child_results:
            true_subset = true_subset.compute_union(child_result.true_subset)

        return AutomationResult(context, true_subset, child_results=child_results)

    @public
    def replace(
        self, old: Union[AutomationCondition, str], new: AutomationCondition
    ) -> AutomationCondition:
        """Replaces all instances of ``old`` across any sub-conditions with ``new``.

        If ``old`` is a string, then conditions with a label matching
        that string will be replaced.

        Args:
            old (Union[AutomationCondition, str]): The condition to replace.
            new (AutomationCondition): The condition to replace with.
        """
        return (
            new
            if old in [self, self.get_label()]
            else copy(self, operands=[child.replace(old, new) for child in self.operands])
        )

    @public
    def allow(self, selection: "AssetSelection") -> "OrAutomationCondition":
        """Applies the ``.allow()`` method across all sub-conditions.

        This impacts any dep-related sub-conditions.

        Args:
            selection (AssetSelection): The selection to allow.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        return copy(
            self,
            operands=[
                child.allow(selection) if _has_allow_ignore(child) else child
                for child in self.operands
            ],
        )

    @public
    def ignore(self, selection: "AssetSelection") -> "OrAutomationCondition":
        """Applies the ``.ignore()`` method across all sub-conditions.

        This impacts any dep-related sub-conditions.

        Args:
            selection (AssetSelection): The selection to ignore.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        return copy(
            self,
            operands=[
                child.ignore(selection) if _has_allow_ignore(child) else child
                for child in self.operands
            ],
        )


@whitelist_for_serdes(storage_name="NotAssetCondition")
@record
class NotAutomationCondition(BuiltinAutomationCondition[T_EntityKey]):
    """This class represents the condition that none of its children evaluate to true."""

    operand: AutomationCondition[T_EntityKey]

    @property
    def description(self) -> str:
        return "Not"

    @property
    def name(self) -> str:
        return "NOT"

    @property
    def children(self) -> Sequence[AutomationCondition[T_EntityKey]]:
        return [self.operand]

    async def evaluate(
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        child_result = await context.for_child_condition(
            child_condition=self.operand, child_index=0, candidate_subset=context.candidate_subset
        ).evaluate_async()
        true_subset = context.candidate_subset.compute_difference(child_result.true_subset)

        return AutomationResult(context, true_subset, child_results=[child_result])

    @public
    def replace(
        self, old: Union[AutomationCondition, str], new: AutomationCondition
    ) -> AutomationCondition:
        """Replaces all instances of ``old`` across any sub-conditions with ``new``.

        If ``old`` is a string, then conditions with a label matching
        that string will be replaced.

        Args:
            old (Union[AutomationCondition, str]): The condition to replace.
            new (AutomationCondition): The condition to replace with.
        """
        return (
            new
            if old in [self, self.get_label()]
            else copy(self, operand=self.operand.replace(old, new))
        )

    @public
    def allow(self, selection: "AssetSelection") -> "NotAutomationCondition":
        """Applies the ``.allow()`` method across all sub-conditions.

        This impacts any dep-related sub-conditions.

        Args:
            selection (AssetSelection): The selection to allow.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        return copy(
            self,
            operand=self.operand.allow(selection)
            if _has_allow_ignore(self.operand)
            else self.operand,
        )

    @public
    def ignore(self, selection: "AssetSelection") -> "NotAutomationCondition":
        """Applies the ``.ignore()`` method across all sub-conditions.

        This impacts any dep-related sub-conditions.

        Args:
            selection (AssetSelection): The selection to ignore.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        return copy(
            self,
            operand=self.operand.ignore(selection)
            if _has_allow_ignore(self.operand)
            else self.operand,
        )



import warnings
from collections.abc import Mapping, Sequence
from datetime import datetime
from itertools import groupby
from typing import TYPE_CHECKING, AbstractSet, Any, NamedTuple, Optional, Union  # noqa: UP035

import dagster._check as check
from dagster._annotations import deprecated, deprecated_param
from dagster._core.definitions import AssetKey
from dagster._core.definitions.asset_job import build_asset_job, get_asset_graph_for_job
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.backfill_policy import resolve_backfill_policy
from dagster._core.definitions.config import ConfigMapping
from dagster._core.definitions.executor_definition import ExecutorDefinition
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.definitions.metadata import RawMetadataValue
from dagster._core.definitions.partition import PartitionedConfig, PartitionsDefinition
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.run_request import RunRequest
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.instance import DynamicPartitionsStore
from dagster._utils.tags import normalize_tags

if TYPE_CHECKING:
    from dagster._core.definitions import JobDefinition
    from dagster._core.definitions.asset_graph import AssetGraph
    from dagster._core.definitions.asset_selection import CoercibleToAssetSelection
    from dagster._core.definitions.run_config import RunConfig


class UnresolvedAssetJobDefinition(
    NamedTuple(
        "_UnresolvedAssetJobDefinition",
        [
            ("name", str),
            ("selection", AssetSelection),
            (
                "config",
                Optional[Union[ConfigMapping, Mapping[str, Any], "PartitionedConfig"]],
            ),
            ("description", Optional[str]),
            ("tags", Optional[Mapping[str, str]]),
            ("run_tags", Optional[Mapping[str, str]]),
            ("metadata", Optional[Mapping[str, RawMetadataValue]]),
            ("partitions_def", Optional[PartitionsDefinition]),
            ("executor_def", Optional[ExecutorDefinition]),
            ("hooks", Optional[AbstractSet[HookDefinition]]),
            ("op_retry_policy", Optional["RetryPolicy"]),
        ],
    )
):
    def __new__(
        cls,
        name: str,
        selection: AssetSelection,
        config: Optional[
            Union[ConfigMapping, Mapping[str, Any], "PartitionedConfig", "RunConfig"]
        ] = None,
        description: Optional[str] = None,
        tags: Optional[Mapping[str, str]] = None,
        run_tags: Optional[Mapping[str, str]] = None,
        metadata: Optional[Mapping[str, RawMetadataValue]] = None,
        partitions_def: Optional[PartitionsDefinition] = None,
        executor_def: Optional[ExecutorDefinition] = None,
        hooks: Optional[AbstractSet[HookDefinition]] = None,
        op_retry_policy: Optional["RetryPolicy"] = None,
    ):
        from dagster._core.definitions import ExecutorDefinition
        from dagster._core.definitions.run_config import convert_config_input

        tags = check.opt_mapping_param(tags, "tags")
        # If `run_tags` is set, then we use it, otherwise `tags` acts as both definition tags and
        # run tags. This is for backcompat with old behavior prior to the introduction of
        # `run_tags`.
        run_tags = check.mapping_param(run_tags, "run_tags") if run_tags else tags
        return super().__new__(
            cls,
            name=check.str_param(name, "name"),
            selection=check.inst_param(selection, "selection", AssetSelection),
            config=convert_config_input(config),
            description=check.opt_str_param(description, "description"),
            tags=tags,
            run_tags=run_tags,
            metadata=check.opt_mapping_param(metadata, "metadata"),
            partitions_def=check.opt_inst_param(
                partitions_def, "partitions_def", PartitionsDefinition
            ),
            executor_def=check.opt_inst_param(executor_def, "executor_def", ExecutorDefinition),
            hooks=check.opt_nullable_set_param(hooks, "hooks", of_type=HookDefinition),
            op_retry_policy=check.opt_inst_param(op_retry_policy, "op_retry_policy", RetryPolicy),
        )

    @deprecated(
        breaking_version="2.0.0",
        additional_warn_text="Directly instantiate `RunRequest(partition_key=...)` instead.",
    )
    def run_request_for_partition(
        self,
        partition_key: str,
        run_key: Optional[str] = None,
        tags: Optional[Mapping[str, str]] = None,
        asset_selection: Optional[Sequence[AssetKey]] = None,
        run_config: Optional[Mapping[str, Any]] = None,
        current_time: Optional[datetime] = None,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> RunRequest:
        """Creates a RunRequest object for a run that processes the given partition.

        Args:
            partition_key: The key of the partition to request a run for.
            run_key (Optional[str]): A string key to identify this launched run. For sensors, ensures that
                only one run is created per run key across all sensor evaluations.  For schedules,
                ensures that one run is created per tick, across failure recoveries. Passing in a `None`
                value means that a run will always be launched per evaluation.
            tags (Optional[Dict[str, str]]): A dictionary of tags (string key-value pairs) to attach
                to the launched run.
            run_config (Optional[Mapping[str, Any]]: Configuration for the run. If the job has
                a :py:class:`PartitionedConfig`, this value will override replace the config
                provided by it.
            current_time (Optional[datetime]): Used to determine which time-partitions exist.
                Defaults to now.
            dynamic_partitions_store (Optional[DynamicPartitionsStore]): The DynamicPartitionsStore
                object that is responsible for fetching dynamic partitions. Required when the
                partitions definition is a DynamicPartitionsDefinition with a name defined. Users
                can pass the DagsterInstance fetched via `context.instance` to this argument.

        Returns:
            RunRequest: an object that requests a run to process the given partition.
        """
        from dagster._core.definitions.partition import (
            DynamicPartitionsDefinition,
            PartitionedConfig,
        )

        if not self.partitions_def:
            check.failed("Called run_request_for_partition on a non-partitioned job")

        partitioned_config = PartitionedConfig.from_flexible_config(
            self.config, self.partitions_def
        )

        if (
            isinstance(self.partitions_def, DynamicPartitionsDefinition)
            and self.partitions_def.name
        ):
            # Do not support using run_request_for_partition with dynamic partitions,
            # since this requires querying the instance once per run request for the
            # existent dynamic partitions
            check.failed(
                "run_request_for_partition is not supported for dynamic partitions. Instead, use"
                " RunRequest(partition_key=...)"
            )

        self.partitions_def.validate_partition_key(
            partition_key,
            current_time=current_time,
            dynamic_partitions_store=dynamic_partitions_store,
        )

        run_config = (
            run_config
            if run_config is not None
            else partitioned_config.get_run_config_for_partition_key(partition_key)
        )
        run_request_tags = {
            **(tags or {}),
            **partitioned_config.get_tags_for_partition_key(partition_key),
        }

        return RunRequest(
            job_name=self.name,
            run_key=run_key,
            run_config=run_config,
            tags=run_request_tags,
            asset_selection=asset_selection,
            partition_key=partition_key,
        )

    def resolve(
        self,
        asset_graph: "AssetGraph",
        default_executor_def: Optional[ExecutorDefinition] = None,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    ) -> "JobDefinition":
        """Resolve this UnresolvedAssetJobDefinition into a JobDefinition."""
        try:
            job_asset_graph = get_asset_graph_for_job(asset_graph, self.selection)
        except DagsterInvalidDefinitionError as e:
            raise DagsterInvalidDefinitionError(
                f'Error resolving selection for asset job "{self.name}": {e}'
            ) from e

        # Require that all assets in the job have the same backfill policy
        executable_nodes = {job_asset_graph.get(k) for k in job_asset_graph.executable_asset_keys}
        nodes_by_backfill_policy = dict(
            groupby((n for n in executable_nodes if n.is_partitioned), lambda n: n.backfill_policy)
        )
        backfill_policy = resolve_backfill_policy(nodes_by_backfill_policy.keys())

        if len(nodes_by_backfill_policy) > 1:
            keys_by_backfill_policy = {
                bp: [n.key.to_user_string() for n in nodes]
                for bp, nodes in nodes_by_backfill_policy.items()
            }
            warnings.warn(
                f"Asset job {self.name} materializes assets with varying BackfillPolicies."
                f" Using backfill policy with minimum `max_partitions_per_run`: {backfill_policy}."
                f"\n\nFound multiple backfill policies:\n\n{keys_by_backfill_policy}"
            )

        # Error if a PartitionedConfig is defined and any target asset has a backfill policy that
        # materializes anything other than a single partition per run. This is because
        # PartitionedConfig is a function that maps single partition keys to run config, so it's
        # behavior is undefined for multiple-partition runs.
        if backfill_policy.max_partitions_per_run != 1 and isinstance(
            self.config, PartitionedConfig
        ):
            raise DagsterInvalidDefinitionError(
                f"Asset job {self.name} materializes an asset with a BackfillPolicy targeting multiple partitions per run,"
                "but a PartitionedConfig was provided. PartitionedConfigs are not supported for "
                "jobs with multi-partition-per-run backfill policies."
            )

        return build_asset_job(
            self.name,
            asset_graph=job_asset_graph,
            config=self.config,
            description=self.description,
            tags=self.tags,
            run_tags=self.run_tags,
            metadata=self.metadata,
            partitions_def=self.partitions_def,
            executor_def=self.executor_def or default_executor_def,
            hooks=self.hooks,
            op_retry_policy=self.op_retry_policy,
            resource_defs=resource_defs,
            allow_different_partitions_defs=False,
        )


@deprecated_param(
    param="partitions_def",
    breaking_version="2.0.0",
    additional_warn_text="Partitioning is inferred from the selected assets, so setting this is redundant.",
)
def define_asset_job(
    name: str,
    selection: Optional["CoercibleToAssetSelection"] = None,
    config: Optional[
        Union[ConfigMapping, Mapping[str, Any], "PartitionedConfig", "RunConfig"]
    ] = None,
    description: Optional[str] = None,
    tags: Optional[Mapping[str, object]] = None,
    run_tags: Optional[Mapping[str, object]] = None,
    metadata: Optional[Mapping[str, RawMetadataValue]] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    executor_def: Optional[ExecutorDefinition] = None,
    hooks: Optional[AbstractSet[HookDefinition]] = None,
    op_retry_policy: Optional["RetryPolicy"] = None,
) -> UnresolvedAssetJobDefinition:
    """Creates a definition of a job which will either materialize a selection of assets or observe
    a selection of source assets. This will only be resolved to a JobDefinition once placed in a
    code location.

    Args:
        name (str):
            The name for the job.
        selection (Union[str, Sequence[str], Sequence[AssetKey], Sequence[Union[AssetsDefinition, SourceAsset]], AssetSelection]):
            The assets that will be materialized or observed when the job is run.

            The selected assets must all be included in the assets that are passed to the assets
            argument of the Definitions object that this job is included on.

            The string "my_asset*" selects my_asset and all downstream assets within the code
            location. A list of strings represents the union of all assets selected by strings
            within the list.

            The selection will be resolved to a set of assets when the location is loaded. If the
            selection resolves to all source assets, the created job will perform source asset
            observations. If the selection resolves to all regular assets, the created job will
            materialize assets. If the selection resolves to a mixed set of source assets and
            regular assets, an error will be thrown.

        config:
            Describes how the Job is parameterized at runtime.

            If no value is provided, then the schema for the job's run config is a standard
            format based on its ops and resources.

            If a dictionary is provided, then it must conform to the standard config schema, and
            it will be used as the job's run config for the job whenever the job is executed.
            The values provided will be viewable and editable in the Dagster UI, so be
            careful with secrets.

            If a :py:class:`ConfigMapping` object is provided, then the schema for the job's run config is
            determined by the config mapping, and the ConfigMapping, which should return
            configuration in the standard format to configure the job.
        tags (Optional[Mapping[str, object]]): A set of key-value tags that annotate the job and can
            be used for searching and filtering in the UI. Values that are not already strings will
            be serialized as JSON. If `run_tags` is not set, then the content of `tags` will also be
            automatically appended to the tags of any runs of this job.
        run_tags (Optional[Mapping[str, object]]):
            A set of key-value tags that will be automatically attached to runs launched by this
            job. Values that are not already strings will be serialized as JSON. These tag values
            may be overwritten by tag values provided at invocation time. If `run_tags` is set, then
            `tags` are not automatically appended to the tags of any runs of this job.
        metadata (Optional[Mapping[str, RawMetadataValue]]): Arbitrary metadata about the job.
            Keys are displayed string labels, and values are one of the following: string, float,
            int, JSON-serializable dict, JSON-serializable list, and one of the data classes
            returned by a MetadataValue static method.
        description (Optional[str]):
            A description for the Job.
        executor_def (Optional[ExecutorDefinition]):
            How this Job will be executed. Defaults to :py:class:`multi_or_in_process_executor`,
            which can be switched between multi-process and in-process modes of execution. The
            default mode of execution is multi-process.
        op_retry_policy (Optional[RetryPolicy]): The default retry policy for all ops that compute assets in this job.
            Only used if retry policy is not defined on the asset definition.
        partitions_def (Optional[PartitionsDefinition]): (Deprecated)
            Defines the set of partitions for this job. Deprecated because partitioning is inferred
            from the selected assets, so setting this is redundant.


    Returns:
        UnresolvedAssetJobDefinition: The job, which can be placed inside a code location.

    Examples:
        .. code-block:: python

            # A job that targets all assets in the code location:
            @asset
            def asset1():
                ...

            defs = Definitions(
                assets=[asset1],
                jobs=[define_asset_job("all_assets")],
            )

            # A job that targets a single asset
            @asset
            def asset1():
                ...

            defs = Definitions(
                assets=[asset1],
                jobs=[define_asset_job("all_assets", selection=[asset1])],
            )

            # A job that targets all the assets in a group:
            defs = Definitions(
                assets=assets,
                jobs=[define_asset_job("marketing_job", selection=AssetSelection.groups("marketing"))],
            )

            @observable_source_asset
            def source_asset():
                ...

            # A job that observes a source asset:
            defs = Definitions(
                assets=assets,
                jobs=[define_asset_job("observation_job", selection=[source_asset])],
            )

            # Resources are supplied to the assets, not the job:
            @asset(required_resource_keys={"slack_client"})
            def asset1():
                ...

            defs = Definitions(
                assets=[asset1],
                jobs=[define_asset_job("all_assets")],
                resources={"slack_client": prod_slack_client},
            )

    """
    from dagster._core.definitions import AssetSelection

    # convert string-based selections to AssetSelection objects
    if selection is None:
        resolved_selection = AssetSelection.all()
    else:
        resolved_selection = AssetSelection.from_coercible(selection)

    return UnresolvedAssetJobDefinition(
        name=name,
        selection=resolved_selection,
        config=config,
        description=description,
        tags=normalize_tags(tags),
        # Need to preserve None value
        run_tags=normalize_tags(run_tags) if run_tags is not None else None,
        metadata=metadata,
        partitions_def=partitions_def,
        executor_def=executor_def,
        hooks=hooks,
        op_retry_policy=op_retry_policy,
    )



from typing import Optional

import dagster._check as check
from dagster._core.remote_representation.job_index import JobIndex
from dagster._core.remote_representation.represented import RepresentedJob
from dagster._core.snap import JobSnap


class HistoricalJob(RepresentedJob):
    """HistoricalPipeline represents a pipeline that executed in the past
    and has been reloaded into process by querying the instance. Notably
    the user must pass in the pipeline snapshot id that was originally
    assigned to the snapshot, rather than recomputing it which could
    end up different if the schema of the snapshot has changed
    since persistence.
    """

    def __init__(
        self,
        job_snapshot: JobSnap,
        identifying_job_snapshot_id: str,
        parent_job_snapshot: Optional[JobSnap],
    ):
        self._snapshot = check.inst_param(job_snapshot, "job_snapshot", JobSnap)
        self._parent_snapshot = check.opt_inst_param(
            parent_job_snapshot, "parent_job_snapshot", JobSnap
        )
        self._identifying_job_snapshot_id = check.str_param(
            identifying_job_snapshot_id, "identifying_job_snapshot_id"
        )
        self._index = None

    @property
    def _job_index(self):
        if self._index is None:
            self._index = JobIndex(
                self._snapshot,
                self._parent_snapshot,
            )
        return self._index

    @property
    def identifying_job_snapshot_id(self):
        return self._identifying_job_snapshot_id

    @property
    def computed_job_snapshot_id(self):
        return self._job_index.job_snapshot_id



import os
import sys
from collections.abc import AsyncIterable, AsyncIterator, Iterator, Sequence
from contextlib import asynccontextmanager, contextmanager
from threading import Event
from typing import Any, NoReturn, Optional, cast

import google.protobuf.message
import grpc
from grpc_health.v1 import health_pb2
from grpc_health.v1.health_pb2_grpc import HealthStub

import dagster._check as check
import dagster._seven as seven
from dagster._core.errors import DagsterUserCodeUnreachableError
from dagster._core.events import EngineEventData
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.origin import RemoteRepositoryOrigin
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._grpc.__generated__ import DagsterApiStub, dagster_api_pb2
from dagster._grpc.server import GrpcServerProcess
from dagster._grpc.types import (
    CanCancelExecutionRequest,
    CancelExecutionRequest,
    ExecuteExternalJobArgs,
    ExecutionPlanSnapshotArgs,
    ExternalScheduleExecutionArgs,
    JobSubsetSnapshotArgs,
    PartitionArgs,
    PartitionNamesArgs,
    PartitionSetExecutionParamArgs,
    SensorExecutionArgs,
)
from dagster._grpc.utils import (
    default_grpc_timeout,
    default_repository_grpc_timeout,
    default_schedule_grpc_timeout,
    default_sensor_grpc_timeout,
    max_rx_bytes,
    max_send_bytes,
)
from dagster._serdes import serialize_value
from dagster._utils.error import serializable_error_info_from_exc_info

CLIENT_HEARTBEAT_INTERVAL = 1

DEFAULT_GRPC_TIMEOUT = default_grpc_timeout()
DEFAULT_SCHEDULE_GRPC_TIMEOUT = default_schedule_grpc_timeout()
DEFAULT_SENSOR_GRPC_TIMEOUT = default_sensor_grpc_timeout()
DEFAULT_REPOSITORY_GRPC_TIMEOUT = default_repository_grpc_timeout()


def client_heartbeat_thread(client: "DagsterGrpcClient", shutdown_event: Event) -> None:
    while True:
        shutdown_event.wait(CLIENT_HEARTBEAT_INTERVAL)
        if shutdown_event.is_set():
            break

        try:
            client.heartbeat("ping")
        except DagsterUserCodeUnreachableError:
            continue


class DagsterGrpcClient:
    def __init__(
        self,
        port: Optional[int] = None,
        socket: Optional[str] = None,
        host: str = "localhost",
        use_ssl: bool = False,
        metadata: Optional[Sequence[tuple[str, str]]] = None,
    ):
        self.port = check.opt_int_param(port, "port")

        self.socket = check.opt_str_param(socket, "socket")
        self.host = check.opt_str_param(host, "host")
        self._use_ssl = check.bool_param(use_ssl, "use_ssl")

        self._ssl_creds = grpc.ssl_channel_credentials() if use_ssl else None

        self._metadata = check.opt_sequence_param(metadata, "metadata")

        check.invariant(
            port is not None if seven.IS_WINDOWS else True,
            "You must pass a valid `port` on Windows: `socket` not supported.",
        )
        check.invariant(
            (port or socket) and not (port and socket),
            "You must pass one and only one of `port` or `socket`.",
        )
        check.invariant(
            host is not None if port else True,
            "Must provide a hostname",
        )

        if port:
            self._server_address = host + ":" + str(port)
        else:
            socket = check.not_none(socket)
            self._server_address = "unix:" + os.path.abspath(socket)

    @property
    def metadata(self) -> Sequence[tuple[str, str]]:
        return self._metadata

    @property
    def use_ssl(self) -> bool:
        return self._use_ssl

    @contextmanager
    def _channel(self) -> Iterator[grpc.Channel]:
        options = [
            ("grpc.max_receive_message_length", max_rx_bytes()),
            ("grpc.max_send_message_length", max_send_bytes()),
        ]
        with (
            grpc.secure_channel(
                self._server_address,
                self._ssl_creds,
                options=options,
                compression=grpc.Compression.Gzip,
            )
            if self._use_ssl
            else grpc.insecure_channel(
                self._server_address,
                options=options,
                compression=grpc.Compression.Gzip,
            )
        ) as channel:
            yield channel

    @asynccontextmanager
    async def _async_channel(self) -> AsyncIterator[grpc.Channel]:
        options = [
            ("grpc.max_receive_message_length", max_rx_bytes()),
            ("grpc.max_send_message_length", max_send_bytes()),
        ]
        async with (
            grpc.aio.secure_channel(
                self._server_address,
                self._ssl_creds,
                options=options,
                compression=grpc.Compression.Gzip,
            )
            if self._use_ssl
            else grpc.aio.insecure_channel(
                self._server_address,
                options=options,
                compression=grpc.Compression.Gzip,
            )
        ) as channel:
            yield channel

    def _get_response(
        self,
        method: str,
        request: google.protobuf.message.Message,
        timeout: int = DEFAULT_GRPC_TIMEOUT,
    ):
        with self._channel() as channel:
            stub = DagsterApiStub(channel)
            return getattr(stub, method)(request, metadata=self._metadata, timeout=timeout)

    async def _gen_response(
        self,
        method: str,
        request: google.protobuf.message.Message,
        timeout: int = DEFAULT_GRPC_TIMEOUT,
    ):
        async with self._async_channel() as channel:
            stub = DagsterApiStub(channel)
            return await getattr(stub, method)(request, metadata=self._metadata, timeout=timeout)

    def _raise_grpc_exception(
        self,
        e: Exception,
        timeout: int,
        custom_timeout_message: Optional[str] = None,
    ) -> NoReturn:
        if isinstance(e, grpc.RpcError):
            if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:  # type: ignore  # (bad stubs)
                raise DagsterUserCodeUnreachableError(
                    custom_timeout_message
                    or f"User code server request timed out due to taking longer than {timeout} seconds to complete."
                ) from e
            else:
                raise DagsterUserCodeUnreachableError(
                    f"Could not reach user code server. gRPC Error code: {e.code().name}"  # type: ignore  # (bad stubs)
                ) from e
        else:
            raise DagsterUserCodeUnreachableError("Could not reach user code server") from e

    def _query(
        self,
        method: str,
        request_type: type[google.protobuf.message.Message],
        timeout: int = DEFAULT_GRPC_TIMEOUT,
        custom_timeout_message: Optional[str] = None,
        **kwargs,
    ):
        try:
            return self._get_response(method, request=request_type(**kwargs), timeout=timeout)
        except Exception as e:
            self._raise_grpc_exception(
                e, timeout=timeout, custom_timeout_message=custom_timeout_message
            )

    async def _gen_query(
        self,
        method: str,
        request_type: type[google.protobuf.message.Message],
        timeout: int = DEFAULT_GRPC_TIMEOUT,
        custom_timeout_message: Optional[str] = None,
        **kwargs,
    ):
        try:
            return await self._gen_response(method, request=request_type(**kwargs), timeout=timeout)
        except Exception as e:
            self._raise_grpc_exception(
                e, timeout=timeout, custom_timeout_message=custom_timeout_message
            )

    def _get_streaming_response(
        self,
        method: str,
        request: google.protobuf.message.Message,
        timeout: int = DEFAULT_GRPC_TIMEOUT,
    ) -> Iterator[Any]:
        with self._channel() as channel:
            stub = DagsterApiStub(channel)
            yield from getattr(stub, method)(request, metadata=self._metadata, timeout=timeout)

    async def _gen_streaming_response(
        self,
        method: str,
        request: google.protobuf.message.Message,
        timeout: int = DEFAULT_GRPC_TIMEOUT,
    ) -> AsyncIterator[Any]:
        async with self._async_channel() as channel:
            stub = DagsterApiStub(channel)
            async for response in getattr(stub, method)(
                request, metadata=self._metadata, timeout=timeout
            ):
                yield response

    def _streaming_query(
        self,
        method: str,
        request_type: type[google.protobuf.message.Message],
        timeout=DEFAULT_GRPC_TIMEOUT,
        custom_timeout_message=None,
        **kwargs,
    ) -> Iterator[Any]:
        try:
            yield from self._get_streaming_response(
                method, request=request_type(**kwargs), timeout=timeout
            )
        except Exception as e:
            self._raise_grpc_exception(
                e, timeout=timeout, custom_timeout_message=custom_timeout_message
            )

    async def _gen_streaming_query(
        self,
        method: str,
        request_type: type[google.protobuf.message.Message],
        timeout=DEFAULT_GRPC_TIMEOUT,
        custom_timeout_message=None,
        **kwargs,
    ) -> AsyncIterable[Any]:
        try:
            async for response in self._gen_streaming_response(
                method, request=request_type(**kwargs), timeout=timeout
            ):
                yield response
        except Exception as e:
            self._raise_grpc_exception(
                e, timeout=timeout, custom_timeout_message=custom_timeout_message
            )

    def ping(self, echo: str) -> dict[str, Any]:
        check.str_param(echo, "echo")
        res = self._query("Ping", dagster_api_pb2.PingRequest, echo=echo)
        return {
            "echo": res.echo,
            "serialized_server_utilization_metrics": res.serialized_server_utilization_metrics,
        }

    def heartbeat(self, echo: str = "") -> str:
        check.str_param(echo, "echo")
        res = self._query("Heartbeat", dagster_api_pb2.PingRequest, echo=echo)
        return res.echo

    def streaming_ping(self, sequence_length: int, echo: str) -> Iterator[dict]:
        check.int_param(sequence_length, "sequence_length")
        check.str_param(echo, "echo")

        for res in self._streaming_query(
            "StreamingPing",
            dagster_api_pb2.StreamingPingRequest,
            sequence_length=sequence_length,
            echo=echo,
        ):
            yield {
                "sequence_number": res.sequence_number,
                "echo": res.echo,
            }

    def get_server_id(self, timeout: int = DEFAULT_GRPC_TIMEOUT) -> str:
        res = self._query("GetServerId", dagster_api_pb2.Empty, timeout=timeout)
        return res.server_id

    def execution_plan_snapshot(
        self, execution_plan_snapshot_args: ExecutionPlanSnapshotArgs
    ) -> str:
        check.inst_param(
            execution_plan_snapshot_args, "execution_plan_snapshot_args", ExecutionPlanSnapshotArgs
        )
        res = self._query(
            "ExecutionPlanSnapshot",
            dagster_api_pb2.ExecutionPlanSnapshotRequest,
            serialized_execution_plan_snapshot_args=serialize_value(execution_plan_snapshot_args),
        )
        return res.serialized_execution_plan_snapshot

    def list_repositories(self) -> str:
        res = self._query("ListRepositories", dagster_api_pb2.ListRepositoriesRequest)
        return res.serialized_list_repositories_response_or_error

    async def gen_list_repositories(self, **kwargs) -> str:
        res = await self._gen_query(
            "ListRepositories", dagster_api_pb2.ListRepositoriesRequest, **kwargs
        )
        return res.serialized_list_repositories_response_or_error

    def external_partition_names(self, partition_names_args: PartitionNamesArgs) -> str:
        check.inst_param(partition_names_args, "partition_names_args", PartitionNamesArgs)

        res = self._query(
            "ExternalPartitionNames",
            dagster_api_pb2.ExternalPartitionNamesRequest,
            serialized_partition_names_args=serialize_value(partition_names_args),
        )

        return res.serialized_external_partition_names_or_external_partition_execution_error

    def external_partition_config(self, partition_args: PartitionArgs) -> str:
        check.inst_param(partition_args, "partition_args", PartitionArgs)

        res = self._query(
            "ExternalPartitionConfig",
            dagster_api_pb2.ExternalPartitionConfigRequest,
            serialized_partition_args=serialize_value(partition_args),
        )

        return res.serialized_external_partition_config_or_external_partition_execution_error

    def external_partition_tags(self, partition_args: PartitionArgs) -> str:
        check.inst_param(partition_args, "partition_args", PartitionArgs)

        res = self._query(
            "ExternalPartitionTags",
            dagster_api_pb2.ExternalPartitionTagsRequest,
            serialized_partition_args=serialize_value(partition_args),
        )

        return res.serialized_external_partition_tags_or_external_partition_execution_error

    def external_partition_set_execution_params(
        self, partition_set_execution_param_args: PartitionSetExecutionParamArgs
    ) -> str:
        check.inst_param(
            partition_set_execution_param_args,
            "partition_set_execution_param_args",
            PartitionSetExecutionParamArgs,
        )

        chunks = list(
            self._streaming_query(
                "ExternalPartitionSetExecutionParams",
                dagster_api_pb2.ExternalPartitionSetExecutionParamsRequest,
                serialized_partition_set_execution_param_args=serialize_value(
                    partition_set_execution_param_args
                ),
            )
        )

        return "".join([chunk.serialized_chunk for chunk in chunks])

    def external_pipeline_subset(self, pipeline_subset_snapshot_args: JobSubsetSnapshotArgs) -> str:
        check.inst_param(
            pipeline_subset_snapshot_args,
            "pipeline_subset_snapshot_args",
            JobSubsetSnapshotArgs,
        )

        res = self._query(
            "ExternalPipelineSubsetSnapshot",
            dagster_api_pb2.ExternalPipelineSubsetSnapshotRequest,
            serialized_pipeline_subset_snapshot_args=serialize_value(pipeline_subset_snapshot_args),
        )

        return res.serialized_external_pipeline_subset_result

    def reload_code(self, timeout: int) -> dagster_api_pb2.ReloadCodeReply:
        return self._query("ReloadCode", dagster_api_pb2.ReloadCodeRequest, timeout=timeout)

    def external_repository(
        self,
        remote_repository_origin: RemoteRepositoryOrigin,
        defer_snapshots: bool = False,
    ) -> str:
        check.inst_param(
            remote_repository_origin,
            "remote_repository_origin",
            RemoteRepositoryOrigin,
        )

        res = self._query(
            "ExternalRepository",
            dagster_api_pb2.ExternalRepositoryRequest,
            # rename this param name
            serialized_repository_python_origin=serialize_value(remote_repository_origin),
            defer_snapshots=defer_snapshots,
        )

        return res.serialized_external_repository_data

    def external_job(
        self,
        remote_repository_origin: RemoteRepositoryOrigin,
        job_name: str,
        timeout=DEFAULT_GRPC_TIMEOUT,
    ) -> dagster_api_pb2.ExternalJobReply:
        check.inst_param(
            remote_repository_origin,
            "remote_repository_origin",
            RemoteRepositoryOrigin,
        )

        return self._query(
            "ExternalJob",
            dagster_api_pb2.ExternalJobRequest,
            serialized_repository_origin=serialize_value(remote_repository_origin),
            job_name=job_name,
            timeout=timeout,
        )

    def streaming_external_repository(
        self,
        remote_repository_origin: RemoteRepositoryOrigin,
        defer_snapshots: bool = False,
        timeout=DEFAULT_REPOSITORY_GRPC_TIMEOUT,
    ) -> Iterator[dict]:
        for res in self._streaming_query(
            "StreamingExternalRepository",
            dagster_api_pb2.ExternalRepositoryRequest,
            # Rename parameter
            serialized_repository_python_origin=serialize_value(remote_repository_origin),
            defer_snapshots=defer_snapshots,
            timeout=timeout,
        ):
            yield {
                "sequence_number": res.sequence_number,
                "serialized_external_repository_chunk": res.serialized_external_repository_chunk,
            }

    async def gen_streaming_external_repository(
        self,
        remote_repository_origin: RemoteRepositoryOrigin,
        defer_snapshots: bool = False,
        timeout=DEFAULT_REPOSITORY_GRPC_TIMEOUT,
    ) -> AsyncIterable[dict]:
        async for res in self._gen_streaming_query(
            "StreamingExternalRepository",
            dagster_api_pb2.ExternalRepositoryRequest,
            # Rename parameter
            serialized_repository_python_origin=serialize_value(remote_repository_origin),
            defer_snapshots=defer_snapshots,
            timeout=timeout,
        ):
            yield {
                "sequence_number": res.sequence_number,
                "serialized_external_repository_chunk": res.serialized_external_repository_chunk,
            }

    def _is_unimplemented_error(self, e: Exception) -> bool:
        return (
            isinstance(e.__cause__, grpc.RpcError)
            and cast(grpc.RpcError, e.__cause__).code() == grpc.StatusCode.UNIMPLEMENTED
        )

    def external_schedule_execution(
        self, external_schedule_execution_args: ExternalScheduleExecutionArgs
    ) -> str:
        check.inst_param(
            external_schedule_execution_args,
            "external_schedule_execution_args",
            ExternalScheduleExecutionArgs,
        )

        # The timeout for the schedule can be defined in one of three ways.
        #   1. By the default grpc timeout
        #   2. By the DEFAULT_SCHEDULE_GRPC_TIMEOUT environment variable
        #   3. By the client.
        # The DEFAULT_SCHEDULE_GRPC_TIMEOUT constant takes the maximum of (1) and
        # (2), while
        # the client may pass a timeout argument via the
        # `sensor_execution_args` object. If the timeout is passed from the client, we use that value irrespective of what the other timeout values may be set to.
        timeout = (
            external_schedule_execution_args.timeout
            if external_schedule_execution_args.timeout is not None
            else DEFAULT_SCHEDULE_GRPC_TIMEOUT
        )

        try:
            return self._query(
                "SyncExternalScheduleExecution",
                dagster_api_pb2.ExternalScheduleExecutionRequest,
                serialized_external_schedule_execution_args=serialize_value(
                    external_schedule_execution_args
                ),
                timeout=timeout,
            ).serialized_schedule_result
        except Exception as e:
            # On older servers that only have the streaming API call implemented, fall back to that API
            if self._is_unimplemented_error(e):
                chunks = list(
                    self._streaming_query(
                        "ExternalScheduleExecution",
                        dagster_api_pb2.ExternalScheduleExecutionRequest,
                        serialized_external_schedule_execution_args=serialize_value(
                            external_schedule_execution_args
                        ),
                        timeout=timeout,
                    )
                )
                return "".join([chunk.serialized_chunk for chunk in chunks])
            else:
                raise

    def external_sensor_execution(self, sensor_execution_args: SensorExecutionArgs) -> str:
        check.inst_param(
            sensor_execution_args,
            "sensor_execution_args",
            SensorExecutionArgs,
        )
        # The timeout for the sensor can be defined in one of three ways.
        #   1. By the default grpc timeout
        #   2. By the DEFAULT_SENSOR_GRPC_TIMEOUT environment variable
        #   3. By the client.
        # The DEFAULT_SENSOR_GRPC_TIMEOUT constant takes the maximum of (1) and
        # (2), while
        # the client may pass a timeout argument via the
        # `sensor_execution_args` object. If the timeout is passed from the client, we use that value irrespective of what the other timeout values may be set to.
        timeout = (
            sensor_execution_args.timeout
            if sensor_execution_args.timeout is not None
            else DEFAULT_SENSOR_GRPC_TIMEOUT
        )

        custom_timeout_message = (
            f"The sensor tick timed out due to taking longer than {timeout} seconds to execute the"
            " sensor function. One way to avoid this error is to break up the sensor work into"
            " chunks, using cursors to let subsequent sensor calls pick up where the previous call"
            " left off."
        )

        try:
            return self._query(
                "SyncExternalSensorExecution",
                dagster_api_pb2.ExternalSensorExecutionRequest,
                timeout=timeout,
                serialized_external_sensor_execution_args=serialize_value(sensor_execution_args),
                custom_timeout_message=custom_timeout_message,
            ).serialized_sensor_result
        except Exception as e:
            # On older servers that only have the streaming API call implemented, fall back to that API
            if self._is_unimplemented_error(e):
                chunks = list(
                    self._streaming_query(
                        "ExternalSensorExecution",
                        dagster_api_pb2.ExternalSensorExecutionRequest,
                        timeout=timeout,
                        serialized_external_sensor_execution_args=serialize_value(
                            sensor_execution_args
                        ),
                        custom_timeout_message=custom_timeout_message,
                    )
                )

                return "".join([chunk.serialized_chunk for chunk in chunks])
            else:
                raise

    def external_notebook_data(self, notebook_path: str) -> bytes:
        check.str_param(notebook_path, "notebook_path")
        res = self._query(
            "ExternalNotebookData",
            dagster_api_pb2.ExternalNotebookDataRequest,
            notebook_path=notebook_path,
        )
        return res.content

    def shutdown_server(self, timeout: int = 15) -> str:
        res = self._query("ShutdownServer", dagster_api_pb2.Empty, timeout=timeout)
        return res.serialized_shutdown_server_result

    def cancel_execution(self, cancel_execution_request: CancelExecutionRequest) -> str:
        check.inst_param(
            cancel_execution_request,
            "cancel_execution_request",
            CancelExecutionRequest,
        )

        res = self._query(
            "CancelExecution",
            dagster_api_pb2.CancelExecutionRequest,
            serialized_cancel_execution_request=serialize_value(cancel_execution_request),
        )

        return res.serialized_cancel_execution_result

    def can_cancel_execution(
        self,
        can_cancel_execution_request: CanCancelExecutionRequest,
        timeout: int = DEFAULT_GRPC_TIMEOUT,
    ) -> str:
        check.inst_param(
            can_cancel_execution_request,
            "can_cancel_execution_request",
            CanCancelExecutionRequest,
        )

        res = self._query(
            "CanCancelExecution",
            dagster_api_pb2.CanCancelExecutionRequest,
            timeout=timeout,
            serialized_can_cancel_execution_request=serialize_value(can_cancel_execution_request),
        )

        return res.serialized_can_cancel_execution_result

    def start_run(self, execute_run_args: ExecuteExternalJobArgs) -> str:
        check.inst_param(execute_run_args, "execute_run_args", ExecuteExternalJobArgs)

        with DagsterInstance.from_ref(execute_run_args.instance_ref) as instance:  # type: ignore # (possible none)
            try:
                res = self._query(
                    "StartRun",
                    dagster_api_pb2.StartRunRequest,
                    serialized_execute_run_args=serialize_value(execute_run_args),
                )
                return res.serialized_start_run_result

            except Exception:
                dagster_run = instance.get_run_by_id(execute_run_args.run_id)
                instance.report_engine_event(
                    message="Unexpected error in IPC client",
                    dagster_run=dagster_run,
                    engine_event_data=EngineEventData.engine_error(
                        serializable_error_info_from_exc_info(sys.exc_info())
                    ),
                )
                raise

    def get_current_image(self) -> str:
        res = self._query("GetCurrentImage", dagster_api_pb2.Empty)
        return res.serialized_current_image

    def get_current_runs(self) -> str:
        res = self._query("GetCurrentRuns", dagster_api_pb2.Empty)
        return res.serialized_current_runs

    def health_check_query(self):
        try:
            with self._channel() as channel:
                response = HealthStub(channel).Check(
                    health_pb2.HealthCheckRequest(service="DagsterApi")
                )
        except grpc.RpcError as e:
            print(e)  # noqa: T201
            return health_pb2.HealthCheckResponse.UNKNOWN

        status_number = response.status

        return health_pb2.HealthCheckResponse.ServingStatus.Name(status_number)


@contextmanager
def ephemeral_grpc_api_client(
    loadable_target_origin: Optional[LoadableTargetOrigin] = None,
    force_port: bool = False,
    max_retries: int = 10,
    max_workers: Optional[int] = None,
) -> Iterator[DagsterGrpcClient]:
    check.opt_inst_param(loadable_target_origin, "loadable_target_origin", LoadableTargetOrigin)
    check.bool_param(force_port, "force_port")
    check.int_param(max_retries, "max_retries")

    from dagster._core.test_utils import instance_for_test

    with instance_for_test() as instance:
        with GrpcServerProcess(
            instance_ref=instance.get_ref(),
            loadable_target_origin=loadable_target_origin,
            force_port=force_port,
            max_retries=max_retries,
            max_workers=max_workers,
            wait_on_exit=True,
        ) as server_process:
            yield server_process.create_client()



"""Core Dagster error classes.

All errors thrown by the Dagster framework inherit from :py:class:`~dagster.DagsterError`. Users
should not subclass this base class for their own exceptions.

There is another exception base class, :py:class:`~dagster.DagsterUserCodeExecutionError`, which is
used by the framework in concert with the :py:func:`~dagster._core.errors.user_code_error_boundary`.

Dagster uses this construct to wrap user code into which it calls. User code can perform arbitrary
computations and may itself throw exceptions. The error boundary catches these user code-generated
exceptions, and then reraises them wrapped in a subclass of
:py:class:`~dagster.DagsterUserCodeExecutionError`.

The wrapped exceptions include additional context for the original exceptions, injected by the
Dagster runtime.
"""

import sys
from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Optional

import dagster._check as check
from dagster._utils.interrupts import raise_interrupts_as

if TYPE_CHECKING:
    from dagster._core.log_manager import DagsterLogManager


class DagsterExecutionInterruptedError(BaseException):
    """Pipeline execution was interrupted during the execution process.

    Just like KeyboardInterrupt this inherits from BaseException
    as to not be accidentally caught by code that catches Exception
    and thus prevent the interpreter from exiting.
    """


class DagsterError(Exception):
    """Base class for all errors thrown by the Dagster framework.

    Users should not subclass this base class for their own exceptions.
    """

    @property
    def is_user_code_error(self):
        """Returns true if this error is attributable to user code."""
        return False


class DagsterInvalidDefinitionError(DagsterError):
    """Indicates that the rules for a definition have been violated by the user."""


class DagsterInvalidObservationError(DagsterError):
    """Indicates that an invalid value was returned from a source asset observation function."""


class DagsterInvalidSubsetError(DagsterError):
    """Indicates that a subset of a pipeline is invalid because either:
    - One or more ops in the specified subset do not exist on the job.'
    - The subset produces an invalid job.
    """


class DagsterInvalidDeserializationVersionError(DagsterError):
    """Indicates that a serialized value has an unsupported version and cannot be deserialized."""


PYTHONIC_CONFIG_ERROR_VERBIAGE = """
This config type can be a:
    - Python primitive type
        - int, float, bool, str, list
    - A Python Dict or List type containing other valid types
    - Custom data classes extending dagster.Config
    - A Pydantic discriminated union type (https://docs.pydantic.dev/usage/types/#discriminated-unions-aka-tagged-unions)
"""

PYTHONIC_RESOURCE_ADDITIONAL_TYPES = """

If this config type represents a resource dependency, its annotation must either:
    - Extend dagster.ConfigurableResource, dagster.ConfigurableIOManager, or
    - Be wrapped in a ResourceDependency annotation, e.g. ResourceDependency[{invalid_type_str}]
"""


def _generate_pythonic_config_error_message(
    config_class: Optional[type],
    field_name: Optional[str],
    invalid_type: Any,
    is_resource: bool = False,
) -> str:
    invalid_type_name = getattr(invalid_type, "__name__", "<my type>")
    pythonic_config_error_verbiage = (
        PYTHONIC_CONFIG_ERROR_VERBIAGE + (PYTHONIC_RESOURCE_ADDITIONAL_TYPES if is_resource else "")
    ).format(invalid_type_str=invalid_type_name)

    return (
        """
Error defining Dagster config class{config_class}{field_name}.
Unable to resolve config type {invalid_type} to a supported Dagster config type.

{PYTHONIC_CONFIG_ERROR_VERBIAGE}"""
    ).format(
        config_class=f" {config_class!r}" if config_class else "",
        field_name=f" on field '{field_name}'" if field_name else "",
        invalid_type=repr(invalid_type),
        PYTHONIC_CONFIG_ERROR_VERBIAGE=pythonic_config_error_verbiage,
    )


class DagsterInvalidPythonicConfigDefinitionError(DagsterError):
    """Indicates that you have attempted to construct a Pythonic config or resource class with an invalid value."""

    def __init__(
        self,
        config_class: Optional[type],
        field_name: Optional[str],
        invalid_type: Any,
        is_resource: bool = False,
        **kwargs,
    ):
        self.invalid_type = invalid_type
        self.field_name = field_name
        self.config_class = config_class
        super().__init__(
            _generate_pythonic_config_error_message(
                config_class=config_class,
                field_name=field_name,
                invalid_type=invalid_type,
                is_resource=is_resource,
            ),
            **kwargs,
        )


class DagsterInvalidDagsterTypeInPythonicConfigDefinitionError(DagsterError):
    """Indicates that you have attempted to construct a Pythonic config or resource class with a DagsterType
    annotated field.
    """

    def __init__(
        self,
        config_class_name: str,
        field_name: Optional[str],
        **kwargs,
    ):
        self.field_name = field_name
        super().__init__(
            f"""Error defining Dagster config class '{config_class_name}' on field '{field_name}'. DagsterTypes cannot be used to annotate a config type. DagsterType is meant only for type checking and coercion in op and asset inputs and outputs.
{PYTHONIC_CONFIG_ERROR_VERBIAGE}""",
            **kwargs,
        )


CONFIG_ERROR_VERBIAGE = """
This value can be a:
    - Field
    - Python primitive types that resolve to dagster config types
        - int, float, bool, str, list.
    - A dagster config type: Int, Float, Bool, Array, Optional, Selector, Shape, Permissive, Map
    - A bare python dictionary, which is wrapped in Field(Shape(...)). Any values
      in the dictionary get resolved by the same rules, recursively.
    - A python list with a single entry that can resolve to a type, e.g. [int]
"""


class DagsterInvalidConfigDefinitionError(DagsterError):
    """Indicates that you have attempted to construct a config with an invalid value.

    Acceptable values for config types are any of:
        1. A Python primitive type that resolves to a Dagster config type
            (:py:class:`~python:int`, :py:class:`~python:float`, :py:class:`~python:bool`,
            :py:class:`~python:str`, or :py:class:`~python:list`).

        2. A Dagster config type: :py:data:`~dagster.Int`, :py:data:`~dagster.Float`,
            :py:data:`~dagster.Bool`, :py:data:`~dagster.String`,
            :py:data:`~dagster.StringSource`, :py:data:`~dagster.Any`,
            :py:class:`~dagster.Array`, :py:data:`~dagster.Noneable`, :py:data:`~dagster.Enum`,
            :py:class:`~dagster.Selector`, :py:class:`~dagster.Shape`, or
            :py:class:`~dagster.Permissive`.

        3. A bare python dictionary, which will be automatically wrapped in
            :py:class:`~dagster.Shape`. Values of the dictionary are resolved recursively
            according to the same rules.

        4. A bare python list of length one which itself is config type.
            Becomes :py:class:`Array` with list element as an argument.

        5. An instance of :py:class:`~dagster.Field`.
    """

    def __init__(self, original_root, current_value, stack, reason=None, **kwargs):
        self.original_root = original_root
        self.current_value = current_value
        self.stack = stack
        super().__init__(
            (
                "Error defining config. Original value passed: {original_root}. "
                "{stack_str}{current_value} "
                "cannot be resolved.{reason_str}" + CONFIG_ERROR_VERBIAGE
            ).format(
                original_root=repr(original_root),
                stack_str="Error at stack path :" + ":".join(stack) + ". " if stack else "",
                current_value=repr(current_value),
                reason_str=f" Reason: {reason}." if reason else "",
            ),
            **kwargs,
        )


class DagsterInvariantViolationError(DagsterError):
    """Indicates the user has violated a well-defined invariant that can only be enforced
    at runtime.
    """


class DagsterExecutionStepNotFoundError(DagsterError):
    """Thrown when the user specifies execution step keys that do not exist."""

    def __init__(self, *args, **kwargs):
        self.step_keys = check.list_param(kwargs.pop("step_keys"), "step_keys", str)
        super().__init__(*args, **kwargs)


class DagsterExecutionPlanSnapshotNotFoundError(DagsterError):
    """Thrown when an expected execution plan snapshot could not be found on a PipelineRun."""


class DagsterRunNotFoundError(DagsterError):
    """Thrown when a run cannot be found in run storage."""

    def __init__(self, *args, **kwargs):
        self.invalid_run_id = check.str_param(kwargs.pop("invalid_run_id"), "invalid_run_id")
        super().__init__(*args, **kwargs)


class DagsterStepOutputNotFoundError(DagsterError):
    """Indicates that previous step outputs required for an execution step to proceed are not
    available.
    """

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop("step_key"), "step_key")
        self.output_name = check.str_param(kwargs.pop("output_name"), "output_name")
        super().__init__(*args, **kwargs)


@contextmanager
def raise_execution_interrupts() -> Iterator[None]:
    with raise_interrupts_as(DagsterExecutionInterruptedError):
        yield


@contextmanager
def user_code_error_boundary(
    error_cls: type["DagsterUserCodeExecutionError"],
    msg_fn: Callable[[], str],
    log_manager: Optional["DagsterLogManager"] = None,
    **kwargs: object,
) -> Iterator[None]:
    """Wraps the execution of user-space code in an error boundary. This places a uniform
    policy around any user code invoked by the framework. This ensures that all user
    errors are wrapped in an exception derived from DagsterUserCodeExecutionError,
    and that the original stack trace of the user error is preserved, so that it
    can be reported without confusing framework code in the stack trace, if a
    tool author wishes to do so.

    Examples:
    .. code-block:: python

        with user_code_error_boundary(
            # Pass a class that inherits from DagsterUserCodeExecutionError
            DagsterExecutionStepExecutionError,
            # Pass a function that produces a message
            "Error occurred during step execution"
        ):
            call_user_provided_function()

    """
    check.callable_param(msg_fn, "msg_fn")
    check.class_param(error_cls, "error_cls", superclass=DagsterUserCodeExecutionError)
    from dagster._utils.error import redact_user_stacktrace_if_enabled

    with redact_user_stacktrace_if_enabled(), raise_execution_interrupts():
        if log_manager:
            log_manager.begin_python_log_capture()
        try:
            yield
        except DagsterError as de:
            # The system has thrown an error that is part of the user-framework contract
            raise de
        except Exception as e:
            # An exception has been thrown by user code and computation should cease
            # with the error reported further up the stack
            new_error = error_cls(
                msg_fn(), user_exception=e, original_exc_info=sys.exc_info(), **kwargs
            )
            raise new_error from e
        finally:
            if log_manager:
                log_manager.end_python_log_capture()


class DagsterUserCodeExecutionError(DagsterError):
    """This is the base class for any exception that is meant to wrap an
    :py:class:`~python:Exception` thrown by user code. It wraps that existing user code.
    The ``original_exc_info`` argument to the constructor is meant to be a tuple of the type
    returned by :py:func:`sys.exc_info <python:sys.exc_info>` at the call site of the constructor.

    Users should not subclass this base class for their own exceptions and should instead throw
    freely from user code. User exceptions will be automatically wrapped and rethrown.
    """

    def __init__(self, *args, **kwargs):
        # original_exc_info should be gotten from a sys.exc_info() call at the
        # callsite inside of the exception handler. this will allow consuming
        # code to *re-raise* the user error in it's original format
        # for cleaner error reporting that does not have framework code in it
        user_exception = check.inst_param(kwargs.pop("user_exception"), "user_exception", Exception)
        original_exc_info = check.tuple_param(kwargs.pop("original_exc_info"), "original_exc_info")

        check.invariant(original_exc_info[0] is not None)

        super().__init__(args[0], *args[1:], **kwargs)

        self.user_exception = check.opt_inst_param(user_exception, "user_exception", Exception)
        self.original_exc_info = original_exc_info

    @property
    def is_user_code_error(self) -> bool:
        return True


class DagsterTypeCheckError(DagsterUserCodeExecutionError):
    """Indicates an error in the op type system at runtime. E.g. a op receives an
    unexpected input, or produces an output that does not match the type of the output definition.
    """


class DagsterExecutionLoadInputError(DagsterUserCodeExecutionError):
    """Indicates an error occurred while loading an input for a step."""

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop("step_key"), "step_key")
        self.input_name = check.str_param(kwargs.pop("input_name"), "input_name")
        super().__init__(*args, **kwargs)


class DagsterExecutionHandleOutputError(DagsterUserCodeExecutionError):
    """Indicates an error occurred while handling an output for a step."""

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop("step_key"), "step_key")
        self.output_name = check.str_param(kwargs.pop("output_name"), "output_name")
        super().__init__(*args, **kwargs)


class DagsterExecutionStepExecutionError(DagsterUserCodeExecutionError):
    """Indicates an error occurred while executing the body of an execution step."""

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop("step_key"), "step_key")
        self.op_name = check.str_param(kwargs.pop("op_name"), "op_name")
        self.op_def_name = check.str_param(kwargs.pop("op_def_name"), "op_def_name")
        super().__init__(*args, **kwargs)


class DagsterResourceFunctionError(DagsterUserCodeExecutionError):
    """Indicates an error occurred while executing the body of the ``resource_fn`` in a
    :py:class:`~dagster.ResourceDefinition` during resource initialization.
    """


class DagsterConfigMappingFunctionError(DagsterUserCodeExecutionError):
    """Indicates that an unexpected error occurred while executing the body of a config mapping
    function defined in a :py:class:`~dagster.JobDefinition` or `~dagster.GraphDefinition` during
    config parsing.
    """


class DagsterTypeLoadingError(DagsterUserCodeExecutionError):
    """Indicates that an unexpected error occurred while executing the body of an type load
    function defined in a :py:class:`~dagster.DagsterTypeLoader` during loading of a custom type.
    """


class DagsterUnknownResourceError(DagsterError, AttributeError):
    # inherits from AttributeError as it is raised within a __getattr__ call... used to support
    # object hasattr method
    """Indicates that an unknown resource was accessed in the body of an execution step. May often
    happen by accessing a resource in the compute function of an op without first supplying the
    op with the correct `required_resource_keys` argument.
    """

    def __init__(self, resource_name, *args, **kwargs):
        self.resource_name = check.str_param(resource_name, "resource_name")
        msg = (
            f"Unknown resource `{resource_name}`. Specify `{resource_name}` as a required resource "
            "on the compute / config function that accessed it."
        )
        super().__init__(msg, *args, **kwargs)


class DagsterInvalidInvocationError(DagsterError):
    """Indicates that an error has occurred when an op has been invoked, but before the actual
    core compute has been reached.
    """


class DagsterInvalidConfigError(DagsterError):
    """Thrown when provided config is invalid (does not type check against the relevant config
    schema).
    """

    def __init__(self, preamble, errors, config_value, *args, **kwargs):
        from dagster._config import EvaluationError

        check.str_param(preamble, "preamble")
        self.errors = check.list_param(errors, "errors", of_type=EvaluationError)
        self.config_value = config_value

        error_msg = preamble
        error_messages = []

        for i_error, error in enumerate(self.errors):
            error_messages.append(error.message)
            error_msg += f"\n    Error {i_error + 1}: {error.message}"

        self.message = error_msg
        self.error_messages = error_messages

        super().__init__(error_msg, *args, **kwargs)


class DagsterUnmetExecutorRequirementsError(DagsterError):
    """Indicates the resolved executor is incompatible with the state of other systems
    such as the :py:class:`~dagster._core.instance.DagsterInstance` or system storage configuration.
    """


class DagsterSubprocessError(DagsterError):
    """An exception has occurred in one or more of the child processes dagster manages.
    This error forwards the message and stack trace for all of the collected errors.
    """

    def __init__(self, *args, **kwargs):
        from dagster._utils.error import SerializableErrorInfo

        self.subprocess_error_infos = check.list_param(
            kwargs.pop("subprocess_error_infos"), "subprocess_error_infos", SerializableErrorInfo
        )
        super().__init__(*args, **kwargs)


class DagsterUserCodeUnreachableError(DagsterError):
    """Dagster was unable to reach a user code server to fetch information about user code."""


class DagsterUserCodeProcessError(DagsterError):
    """An exception has occurred in a user code process that the host process raising this error
    was communicating with.
    """

    @staticmethod
    def from_error_info(error_info):
        from dagster._utils.error import SerializableErrorInfo

        check.inst_param(error_info, "error_info", SerializableErrorInfo)
        return DagsterUserCodeProcessError(
            error_info.to_string(), user_code_process_error_infos=[error_info]
        )

    def __init__(self, *args, **kwargs):
        from dagster._utils.error import SerializableErrorInfo

        self.user_code_process_error_infos = check.list_param(
            kwargs.pop("user_code_process_error_infos"),
            "user_code_process_error_infos",
            SerializableErrorInfo,
        )
        super().__init__(*args, **kwargs)


class DagsterMaxRetriesExceededError(DagsterError):
    """Raised when raise_on_error is true, and retries were exceeded, this error should be raised."""

    def __init__(self, *args, **kwargs):
        from dagster._utils.error import SerializableErrorInfo

        self.user_code_process_error_infos = check.list_param(
            kwargs.pop("user_code_process_error_infos"),
            "user_code_process_error_infos",
            SerializableErrorInfo,
        )
        super().__init__(*args, **kwargs)

    @staticmethod
    def from_error_info(error_info):
        from dagster._utils.error import SerializableErrorInfo

        check.inst_param(error_info, "error_info", SerializableErrorInfo)
        return DagsterMaxRetriesExceededError(
            error_info.to_string(), user_code_process_error_infos=[error_info]
        )


class DagsterCodeLocationNotFoundError(DagsterError):
    pass


class DagsterRedactedUserCodeError(DagsterError):
    """Error used to mask user code errors to prevent leaking sensitive information. Contains an error ID that can be
    used to look up the original error in the user code error log.
    """


class DagsterUserCodeLoadError(DagsterUserCodeExecutionError):
    """Errors raised in a user process during the loading of user code."""


class DagsterCodeLocationLoadError(DagsterError):
    def __init__(self, *args, **kwargs):
        from dagster._utils.error import SerializableErrorInfo

        self.load_error_infos = check.list_param(
            kwargs.pop("load_error_infos"),
            "load_error_infos",
            SerializableErrorInfo,
        )
        super().__init__(*args, **kwargs)


class DagsterLaunchFailedError(DagsterError):
    """Indicates an error while attempting to launch a pipeline run."""

    def __init__(self, *args, **kwargs):
        from dagster._utils.error import SerializableErrorInfo

        self.serializable_error_info = check.opt_inst_param(
            kwargs.pop("serializable_error_info", None),
            "serializable_error_info",
            SerializableErrorInfo,
        )
        super().__init__(*args, **kwargs)


class DagsterBackfillFailedError(DagsterError):
    """Indicates an error while attempting to launch a backfill."""

    def __init__(self, *args, **kwargs):
        from dagster._utils.error import SerializableErrorInfo

        self.serializable_error_info = check.opt_inst_param(
            kwargs.pop("serializable_error_info", None),
            "serializable_error_info",
            SerializableErrorInfo,
        )
        super().__init__(*args, **kwargs)


class DagsterRunAlreadyExists(DagsterError):
    """Indicates that a pipeline run already exists in a run storage."""


class DagsterSnapshotDoesNotExist(DagsterError):
    """Indicates you attempted to create a pipeline run with a nonexistent snapshot id."""


class DagsterRunConflict(DagsterError):
    """Indicates that a conflicting pipeline run exists in a run storage."""


class DagsterTypeCheckDidNotPass(DagsterError):
    """Indicates that a type check failed.

    This is raised when ``raise_on_error`` is ``True`` in calls to the synchronous job and
    graph execution APIs (e.g. `graph.execute_in_process()`, `job.execute_in_process()` -- typically
    within a test), and a :py:class:`~dagster.DagsterType`'s type check fails by returning either
    ``False`` or an instance of :py:class:`~dagster.TypeCheck` whose ``success`` member is ``False``.
    """

    def __init__(self, description=None, metadata=None, dagster_type=None):
        from dagster import DagsterType
        from dagster._core.definitions.metadata import normalize_metadata

        super().__init__(description)
        self.description = check.opt_str_param(description, "description")
        self.metadata = normalize_metadata(
            check.opt_mapping_param(metadata, "metadata", key_type=str)
        )
        self.dagster_type = check.opt_inst_param(dagster_type, "dagster_type", DagsterType)


class DagsterAssetCheckFailedError(DagsterError):
    """Indicates than an asset check failed."""


class DagsterEventLogInvalidForRun(DagsterError):
    """Raised when the event logs for a historical run are malformed or invalid."""

    def __init__(self, run_id):
        self.run_id = check.str_param(run_id, "run_id")
        super().__init__(f"Event logs invalid for run id {run_id}")


class ScheduleExecutionError(DagsterUserCodeExecutionError):
    """Errors raised in a user process during the execution of schedule."""


class SensorExecutionError(DagsterUserCodeExecutionError):
    """Errors raised in a user process during the execution of a sensor (or its job)."""


class PartitionExecutionError(DagsterUserCodeExecutionError):
    """Errors raised during the execution of user-provided functions of a partition set schedule."""


class DagsterInvalidAssetKey(DagsterError):
    """Error raised by invalid asset key."""


class DagsterInvalidMetadata(DagsterError):
    """Error raised by invalid metadata parameters."""


class HookExecutionError(DagsterUserCodeExecutionError):
    """Error raised during the execution of a user-defined hook."""


class RunStatusSensorExecutionError(DagsterUserCodeExecutionError):
    """Error raised during the execution of a user-defined run status sensor."""


class FreshnessPolicySensorExecutionError(DagsterUserCodeExecutionError):
    """Error raised during the execution of a user-defined freshness policy sensor."""


class DagsterImportError(DagsterError):
    """Import error raised while importing user-code."""


class JobError(DagsterUserCodeExecutionError):
    """Errors raised during the execution of user-provided functions for a defined Job."""


class DagsterUnknownStepStateError(DagsterError):
    """When job execution completes with steps in an unknown state."""


class DagsterObjectStoreError(DagsterError):
    """Errors during an object store operation."""


class DagsterInvalidPropertyError(DagsterError):
    """Indicates that an invalid property was accessed. May often happen by accessing a property
    that no longer exists after breaking changes.
    """


class DagsterHomeNotSetError(DagsterError):
    """The user has tried to use a command that requires an instance or invoke DagsterInstance.get()
    without setting DAGSTER_HOME env var.
    """


class DagsterUnknownPartitionError(DagsterError):
    """The user has tried to access run config for a partition name that does not exist."""


class DagsterUndefinedDataVersionError(DagsterError):
    """The user attempted to retrieve the most recent logical version for an asset, but no logical version is defined."""


class DagsterAssetBackfillDataLoadError(DagsterError):
    """Indicates that an asset backfill is now unloadable. May happen when (1) a code location containing
    targeted assets is unloadable or (2) and asset or an asset's partitions definition has been removed.
    """


class DagsterDefinitionChangedDeserializationError(DagsterError):
    """Indicates that a stored value can't be deserialized because the definition needed to interpret
    it has changed.
    """


class DagsterPipesExecutionError(DagsterError):
    """Indicates that an error occurred during the execution of an external process."""



import asyncio
from collections.abc import Sequence

from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._record import record
from dagster._serdes.serdes import whitelist_for_serdes


@whitelist_for_serdes
@record
class SinceCondition(BuiltinAutomationCondition[T_EntityKey]):
    trigger_condition: AutomationCondition[T_EntityKey]
    reset_condition: AutomationCondition[T_EntityKey]

    @property
    def name(self) -> str:
        return "SINCE"

    @property
    def children(self) -> Sequence[AutomationCondition[T_EntityKey]]:
        return [self.trigger_condition, self.reset_condition]

    async def evaluate(
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        # must evaluate child condition over the entire subset to avoid missing state transitions
        child_candidate_subset = context.asset_graph_view.get_full_subset(key=context.key)

        # compute result for trigger and reset conditions
        trigger_result, reset_result = await asyncio.gather(
            *[
                context.for_child_condition(
                    self.trigger_condition, child_index=0, candidate_subset=child_candidate_subset
                ).evaluate_async(),
                context.for_child_condition(
                    self.reset_condition, child_index=1, candidate_subset=child_candidate_subset
                ).evaluate_async(),
            ]
        )

        # take the previous subset that this was true for
        true_subset = context.previous_true_subset or context.get_empty_subset()

        # add in any newly true trigger asset partitions
        true_subset = true_subset.compute_union(trigger_result.true_subset)
        # remove any newly true reset asset partitions
        true_subset = true_subset.compute_difference(reset_result.true_subset)

        return AutomationResult(
            context=context, true_subset=true_subset, child_results=[trigger_result, reset_result]
        )



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Optional

import dagster._check as check
import dagster._seven as seven


def format_duration(milliseconds: float) -> str:
    """Given milliseconds, return human readable duration string such as:
    533ms, 2.1s, 4m52s, 34m12s, 1h4m.
    """
    # under 1 ms
    # ex: 0.83ms
    # ex: 8.3ms
    if milliseconds < 10:
        return f"{round(milliseconds, 2)}ms"

    # between 10 ms and 1000 ms
    # ex: 533ms
    if milliseconds < 1000:
        return f"{int(milliseconds)}ms"

    # between one second and one minute
    # ex: 5.6s
    if milliseconds < 1000 * 60:
        seconds = milliseconds / 1000
        return f"{round(seconds, 2)}s"

    # between one minute and 60 minutes
    # 5m42s
    if milliseconds < 1000 * 60 * 60:
        minutes = int(milliseconds // (1000 * 60))
        seconds = int(milliseconds % (1000 * 60) // (1000))
        return f"{minutes}m{seconds}s"

    # Above one hour
    else:
        hours = int(milliseconds // (1000 * 60 * 60))
        minutes = int(milliseconds % (1000 * 60 * 60) // (1000 * 60))
        return f"{hours}h{minutes}m"


class TimerResult:
    start_time: float
    end_time: Optional[float]

    def __init__(self):
        self.start_time = seven.time_fn()
        self.end_time = None

    @property
    def seconds(self) -> float:
        if self.end_time is None:
            check.failed("end time is not set")
        return self.end_time - self.start_time

    @property
    def millis(self) -> float:
        return self.seconds * 1000


@contextmanager
def time_execution_scope() -> Iterator[TimerResult]:
    """Context manager that times the execution of a block of code.

    Example:
    ..code-block::
        from solid_util.timing import time_execution_scope
        with time_execution_scope() as timer_result:
            do_some_operation()

        print(
            'do_some_operation took {timer_result.millis} milliseconds'.format(
                timer_result=timer_result
            )
        )
    """
    timer_result = TimerResult()
    yield timer_result
    timer_result.end_time = seven.time_fn()



"""add column concurrency default limit

Revision ID: 7e2f3204cf8e
Revises: 6b7fb194ff9c
Create Date: 2025-01-13 15:19:19.331752

"""

import sqlalchemy as sa
from alembic import op
from dagster._core.storage.migration.utils import has_column, has_table

# revision identifiers, used by Alembic.
revision = "7e2f3204cf8e"
down_revision = "6b7fb194ff9c"
branch_labels = None
depends_on = None


def upgrade():
    if has_table("concurrency_limits"):
        if not has_column("concurrency_limits", "using_default_limit"):
            op.add_column(
                "concurrency_limits",
                sa.Column(
                    "using_default_limit",
                    sa.Boolean(),
                    nullable=False,
                    default=False,
                    server_default=sa.false(),
                ),
            )


def downgrade():
    if has_table("concurrency_limits"):
        if has_column("concurrency_limits", "using_default_limit"):
            op.drop_column("concurrency_limits", "using_default_limit")



import datetime
from collections.abc import Iterator, Mapping, Sequence
from typing import Callable, Optional, Union

from dagster import _check as check
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import (
    AssetCheckKey,
    AssetCheckSeverity,
    AssetCheckSpec,
)
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.assets import (
    AssetsDefinition,
    SourceAsset,
    unique_id_from_asset_and_check_keys,
)
from dagster._core.definitions.decorators.asset_check_decorator import (
    MultiAssetCheckFunction,
    multi_asset_check,
)
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.metadata import JsonMetadataValue
from dagster._core.definitions.metadata.metadata_value import TimestampMetadataValue
from dagster._core.event_api import AssetRecordsFilter, EventLogRecord
from dagster._core.events import DagsterEventType
from dagster._core.instance import DagsterInstance

# Constants
DEFAULT_FRESHNESS_SEVERITY = AssetCheckSeverity.WARN
DEFAULT_FRESHNESS_TIMEZONE = "UTC"

# Top-level metadata keys
LAST_UPDATED_TIMESTAMP_METADATA_KEY = "dagster/last_updated_timestamp"
FRESHNESS_PARAMS_METADATA_KEY = "dagster/freshness_params"
# The latest asset record should be no earlier than this timestamp.
LOWER_BOUND_TIMESTAMP_METADATA_KEY = "dagster/freshness_lower_bound_timestamp"
# When an asset is fresh, this represents the timestamp when the asset can become stale again.
FRESH_UNTIL_METADATA_KEY = "dagster/fresh_until_timestamp"
# If this is cron-based freshness, this is the latest tick of the cron.
LATEST_CRON_TICK_METADATA_KEY = "dagster/latest_cron_tick_timestamp"

# dagster/freshness_params inner keys
LOWER_BOUND_DELTA_PARAM_KEY = "lower_bound_delta_seconds"
DEADLINE_CRON_PARAM_KEY = "deadline_cron"
TIMEZONE_PARAM_KEY = "timezone"


def ensure_no_duplicate_assets(
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
) -> None:
    """Finds duplicate assets in the provided list of assets, and errors if any are present.

    Args:
        assets (Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]]): The assets to check for duplicates.

    Returns:
        Sequence[AssetKey]: A list of the duplicate assets.
    """
    asset_keys = assets_to_keys(assets)
    duplicate_assets = [asset_key for asset_key in asset_keys if asset_keys.count(asset_key) > 1]
    check.invariant(
        len(duplicate_assets) == 0,
        f"Found duplicate assets in the provided list of assets: {duplicate_assets}. Please ensure that each asset is unique.",
    )


def asset_to_keys_iterable(
    asset: Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset],
) -> Iterator[AssetKey]:
    if isinstance(asset, AssetsDefinition):
        yield from asset.keys
    elif isinstance(asset, SourceAsset):
        yield asset.key
    else:
        yield AssetKey.from_coercible_or_definition(asset)


def assets_to_keys(
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
) -> Sequence[AssetKey]:
    """Converts the provided assets to a sequence of their contained AssetKeys."""
    return [asset_key for asset in assets for asset_key in asset_to_keys_iterable(asset)]


def ensure_no_duplicate_asset_checks(
    asset_checks: Sequence[AssetChecksDefinition],
) -> None:
    asset_check_keys = [
        asset_check_key
        for asset_check in asset_checks
        for asset_check_key in asset_check.check_keys
    ]
    duplicate_asset_checks = [
        asset_check_key
        for asset_check_key in asset_check_keys
        if asset_check_keys.count(asset_check_key) > 1
    ]
    check.invariant(
        len(duplicate_asset_checks) == 0,
        f"Found duplicate asset checks in the provided list of asset checks: {duplicate_asset_checks}. Please ensure that each provided asset check is unique.",
    )


def retrieve_last_update_record(
    instance: DagsterInstance,
    asset_key: AssetKey,
    partition_key: Optional[str],
) -> Optional[EventLogRecord]:
    """Retrieve the latest materialization or observation record for the given asset.

    If the asset is partitioned, the latest record for the latest partition will be returned.

    There are a few weird edge cases to consider here:
        - Note that this may not be the latest record overall. We only look for the latest materialization
        and the latest observation, and return the latest of those two records that we
        can successfully parse. Take for example, the scenario where we have a materialization, and two
        obserations after that. If the latest observation is not parseable, we will return the materialization,
        even though there might be a more recent observation that we can parse. The reason we do
        this is to avoid an expensive query into the long-tail history of the asset. Retrieving the
        latest record of a particular type is cheap. Retrieving the latest N records is not.
        - If the asset has never been materialized and the latest observation is missing
        `dagster/last_updated_timestamp` metadata, we will return None.

    """
    materializations = instance.fetch_materializations(
        records_filter=AssetRecordsFilter(
            asset_key=asset_key, asset_partitions=[partition_key] if partition_key else None
        ),
        limit=1,
    )
    observations = instance.fetch_observations(
        records_filter=AssetRecordsFilter(
            asset_key=asset_key, asset_partitions=[partition_key] if partition_key else None
        ),
        limit=1,
    )
    if materializations.records and observations.records:
        mat_timestamp = retrieve_timestamp_from_record(materializations.records[0])
        obs_timestamp = retrieve_timestamp_from_record(observations.records[0])
        if not obs_timestamp:
            return materializations.records[0]
        if not mat_timestamp:
            return observations.records[0]
        return (
            materializations.records[0]
            if mat_timestamp > obs_timestamp
            else observations.records[0]
        )
    else:
        return (
            materializations.records[0]
            if materializations.records
            else observations.records[0]
            if observations.records
            else None
        )


def retrieve_timestamp_from_record(asset_record: EventLogRecord) -> Optional[float]:
    """Retrieve the timestamp from the given materialization or observation record."""
    check.inst_param(asset_record, "asset_record", EventLogRecord)
    default_value = None if asset_record.asset_observation else asset_record.timestamp
    return (
        get_timestamp_out_of_metadata_if_set(
            check.not_none(asset_record.asset_event).metadata, LAST_UPDATED_TIMESTAMP_METADATA_KEY
        )
        or default_value
    )


def get_timestamp_out_of_metadata_if_set(metadata: Mapping, key: str) -> Optional[float]:
    if key not in metadata:
        return None
    value = check.inst(
        metadata[key],
        TimestampMetadataValue,
        f"Expected {key} in asset metadat to be a TimestampMetadataValue.",
    )
    return value.value


def get_description_for_freshness_check_result(
    passed: bool,
    update_timestamp: Optional[float],
    last_update_time_lower_bound: datetime.datetime,
    current_timestamp: float,
    expected_partition_key: Optional[str],
    record_arrival_timestamp: Optional[float],
    event_type: Optional[DagsterEventType],
) -> str:
    check.invariant(
        (passed and update_timestamp is not None) or not passed,
        "Should not be possible for check to pass without an update to the asset.",
    )
    out_of_date_observation = record_arrival_timestamp is None or (
        record_arrival_timestamp < last_update_time_lower_bound.timestamp()
        and event_type == DagsterEventType.ASSET_OBSERVATION
    )
    update_time_delta_str = (
        seconds_in_words(current_timestamp - update_timestamp) if update_timestamp else None
    )
    last_update_time_lower_bound_delta_str = seconds_in_words(
        current_timestamp - last_update_time_lower_bound.timestamp()
    )
    return (
        f"Partition {expected_partition_key} is fresh. Expected the partition to arrive within the last {last_update_time_lower_bound_delta_str}, and it arrived {update_time_delta_str} ago."
        if passed and expected_partition_key
        else f"Partition {expected_partition_key} is overdue. Expected the partition to arrive within the last {last_update_time_lower_bound_delta_str}."
        if not passed
        and expected_partition_key  # Since we search for a specific partition, if that partition has never been materialized, we will have no record and is therefore in an "unknown" state, but we can't distinguish between that and an overdue state, so we call it overdue regardless.
        else f"Asset is fresh. Expected an update within the last {last_update_time_lower_bound_delta_str}, and found an update {update_time_delta_str} ago."
        if passed and update_timestamp
        else f"Asset is in an unknown state. Expected an update within the last {last_update_time_lower_bound_delta_str}, but we have not received an observation in that time, so we can't determine the state of the asset."
        if not passed and update_timestamp and out_of_date_observation
        else f"Asset is overdue. Expected an update within the last {last_update_time_lower_bound_delta_str}."
    )


def seconds_in_words(delta: float) -> str:
    """Converts the provided number of seconds to a human-readable string.

    Return format is "X days, Y hours, Z minutes, A seconds".
    """
    # Edge case: If the delta is 0 seconds exactly, we should return "0 seconds".
    if delta == 0:
        return "0 seconds"
    days = int(delta // 86400)
    days_unit = "days" if days > 1 else "day" if days == 1 else None
    hours = int(delta % 86400 // 3600)
    hours_unit = "hours" if hours > 1 else "hour" if hours == 1 else None
    minutes = int((delta % 3600) // 60)
    minutes_unit = "minutes" if minutes > 1 else "minute" if minutes == 1 else None
    seconds = int(delta % 60)
    seconds_unit = "seconds" if seconds > 1 else "second" if seconds == 1 else None
    return ", ".join(
        filter(
            None,
            [
                f"{days} {days_unit}" if days_unit else None,
                f"{hours} {hours_unit}" if hours_unit else None,
                f"{minutes} {minutes_unit}" if minutes_unit else None,
                f"{seconds} {seconds_unit}" if seconds_unit else None,
            ],
        )
    )


def freshness_multi_asset_check(params_metadata: JsonMetadataValue, asset_keys: Sequence[AssetKey]):
    def inner(fn: MultiAssetCheckFunction) -> AssetChecksDefinition:
        return multi_asset_check(
            specs=[
                AssetCheckSpec(
                    "freshness_check",
                    asset=asset_key,
                    metadata={FRESHNESS_PARAMS_METADATA_KEY: params_metadata},
                    description="Evaluates freshness for targeted asset.",
                )
                for asset_key in asset_keys
            ],
            can_subset=True,
            name=f"freshness_check_{unique_id_from_asset_and_check_keys(asset_keys)}",
        )(fn)

    return inner


def build_multi_asset_check(
    check_specs: Sequence[AssetCheckSpec],
    check_fn: Callable[[DagsterInstance, AssetCheckKey], tuple[bool, str]],
    severity: AssetCheckSeverity,
) -> Sequence[AssetChecksDefinition]:
    @multi_asset_check(
        specs=check_specs,
        can_subset=True,
        name=f"asset_check_{unique_id_from_asset_and_check_keys([spec.key for spec in check_specs])}",
    )
    def _checks(context):
        for asset_check_key in context.selected_asset_check_keys:
            passed, description = check_fn(context.instance, asset_check_key)
            yield AssetCheckResult(
                passed=passed,
                description=description,
                severity=severity,
                check_name=asset_check_key.name,
                asset_key=asset_check_key.asset_key,
            )

    return [_checks]



"""add bulk_actions job_name column and backfill_tags table

Revision ID: 16e3655b4d9b
Revises: 1aca709bba64
Create Date: 2024-10-23 13:13:13.390846

"""

import sqlalchemy as sa
from alembic import op
from dagster._core.storage.migration.utils import has_column, has_index, has_table
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision = "16e3655b4d9b"
down_revision = "1aca709bba64"
branch_labels = None
depends_on = None


def upgrade():
    if has_table("bulk_actions"):
        if not has_column("bulk_actions", "job_name"):
            op.add_column("bulk_actions", sa.Column("job_name", sa.Text(), nullable=True))

    if not has_table("backfill_tags"):
        op.create_table(
            "backfill_tags",
            sa.Column(
                "id",
                sa.BigInteger().with_variant(sqlite.INTEGER(), "sqlite"),
                primary_key=True,
                autoincrement=True,
            ),
            sa.Column("backfill_id", sa.String(length=255), nullable=True),
            sa.Column("key", sa.Text(), nullable=True),
            sa.Column("value", sa.Text(), nullable=True),
            sa.PrimaryKeyConstraint("id"),
        )
        op.create_index(
            "idx_backfill_tags_backfill_id",
            "backfill_tags",
            ["backfill_id", "id"],
            unique=False,
            postgresql_concurrently=True,
        )


def downgrade():
    if has_table("bulk_actions"):
        if has_column("bulk_actions", "job_name"):
            op.drop_column("bulk_actions", "job_name")

    if has_table("backfill_tags"):
        if has_index("backfill_tags", "idx_backfill_tags_backfill_id"):
            op.drop_index(
                "idx_backfill_tags_backfill_id",
                "backfill_tags",
                postgresql_concurrently=True,
            )
        op.drop_table("backfill_tags")



from collections.abc import Generator, Mapping
from contextlib import contextmanager
from typing import Any, Optional, cast

import dagster._check as check
from dagster._config import process_config
from dagster._core.definitions.resource_definition import (
    ResourceDefinition,
    Resources,
    ScopedResourcesBuilder,
)
from dagster._core.definitions.run_config import define_resource_dictionary_cls
from dagster._core.errors import DagsterInvalidConfigError
from dagster._core.execution.api import ephemeral_instance_if_missing
from dagster._core.execution.context_creation_job import initialize_console_manager
from dagster._core.execution.resources_init import resource_initialization_manager
from dagster._core.instance import DagsterInstance
from dagster._core.log_manager import DagsterLogManager
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.storage.io_manager import IOManager, IOManagerDefinition
from dagster._core.system_config.objects import ResourceConfig, config_map_resources


def get_mapped_resource_config(
    resource_defs: Mapping[str, ResourceDefinition], resource_config: Mapping[str, Any]
) -> Mapping[str, ResourceConfig]:
    resource_config_schema = define_resource_dictionary_cls(
        resource_defs, set(resource_defs.keys())
    )
    config_evr = process_config(resource_config_schema, resource_config)
    if not config_evr.success:
        raise DagsterInvalidConfigError(
            "Error in config for resources ",
            config_evr.errors,
            resource_config,
        )
    config_value = cast(dict[str, Any], config_evr.value)
    return config_map_resources(resource_defs, config_value)


@contextmanager
def build_resources(
    resources: Mapping[str, Any],
    instance: Optional[DagsterInstance] = None,
    resource_config: Optional[Mapping[str, Any]] = None,
    dagster_run: Optional[DagsterRun] = None,
    log_manager: Optional[DagsterLogManager] = None,
) -> Generator[Resources, None, None]:
    """Context manager that yields resources using provided resource definitions and run config.

    This API allows for using resources in an independent context. Resources will be initialized
    with the provided run config, and optionally, dagster_run. The resulting resources will be
    yielded on a dictionary keyed identically to that provided for `resource_defs`. Upon exiting the
    context, resources will also be torn down safely.

    Args:
        resources (Mapping[str, Any]): Resource instances or definitions to build. All
            required resource dependencies to a given resource must be contained within this
            dictionary, or the resource build will fail.
        instance (Optional[DagsterInstance]): The dagster instance configured to instantiate
            resources on.
        resource_config (Optional[Mapping[str, Any]]): A dict representing the config to be
            provided to each resource during initialization and teardown.
        dagster_run (Optional[PipelineRun]): The pipeline run to provide during resource
            initialization and teardown. If the provided resources require either the `dagster_run`
            or `run_id` attributes of the provided context during resource initialization and/or
            teardown, this must be provided, or initialization will fail.
        log_manager (Optional[DagsterLogManager]): Log Manager to use during resource
            initialization. Defaults to system log manager.

    Examples:
        .. code-block:: python

            from dagster import resource, build_resources

            @resource
            def the_resource():
                return "foo"

            with build_resources(resources={"from_def": the_resource, "from_val": "bar"}) as resources:
                assert resources.from_def == "foo"
                assert resources.from_val == "bar"

    """
    resources = check.mapping_param(resources, "resource_defs", key_type=str)
    instance = check.opt_inst_param(instance, "instance", DagsterInstance)
    resource_config = check.opt_mapping_param(resource_config, "resource_config", key_type=str)
    log_manager = check.opt_inst_param(log_manager, "log_manager", DagsterLogManager)
    resource_defs = wrap_resources_for_execution(resources)
    mapped_resource_config = get_mapped_resource_config(resource_defs, resource_config)

    with ephemeral_instance_if_missing(instance) as dagster_instance:
        resources_manager = resource_initialization_manager(
            resource_defs=resource_defs,
            resource_configs=mapped_resource_config,
            log_manager=log_manager if log_manager else initialize_console_manager(dagster_run),
            execution_plan=None,
            dagster_run=dagster_run,
            resource_keys_to_init=set(resource_defs.keys()),
            instance=dagster_instance,
            emit_persistent_events=False,
        )
        try:
            list(resources_manager.generate_setup_events())
            instantiated_resources = check.inst(
                resources_manager.get_object(), ScopedResourcesBuilder
            )
            yield instantiated_resources.build(
                set(instantiated_resources.resource_instance_dict.keys())
            )
        finally:
            list(resources_manager.generate_teardown_events())


def wrap_resources_for_execution(
    resources: Optional[Mapping[str, Any]] = None,
) -> dict[str, ResourceDefinition]:
    return (
        {
            resource_key: wrap_resource_for_execution(resource)
            for resource_key, resource in resources.items()
        }
        if resources
        else {}
    )


def wrap_resource_for_execution(resource: Any) -> ResourceDefinition:
    from dagster._config.pythonic_config import ConfigurableResourceFactory, PartialResource

    # Wrap instantiated resource values in a resource definition.
    # If an instantiated IO manager is provided, wrap it in an IO manager definition.
    if isinstance(resource, (ConfigurableResourceFactory, PartialResource)):
        return resource.get_resource_definition()
    elif isinstance(resource, ResourceDefinition):
        return resource
    elif isinstance(resource, IOManager):
        return IOManagerDefinition.hardcoded_io_manager(resource)
    else:
        return ResourceDefinition.hardcoded_resource(resource)
